/***************************************************************
*																					*
*	exec.c: Event Processing Executive									*
*																					*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************

							*** HISTORY ***

	060830 ALP: Modified to use linked list for messages and
		events.

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "exec.h"
#include "app.h"

#ifdef DEBUG
#include "cdebug.h"
#endif

/***************************************************************/

//		*** MACROS ****

/***************************************************************/

//		*** GENERAL DEFINITIONS ****


/***************************************************************/

//		*** DATA STRUCTURE Definitions ***

typedef struct TAG_MSG_LIST
{
	struct TAG_MSG *pFirst;	// Pointer to first message in list
	struct TAG_MSG *pLast;	// Pointer to last message in list
//	byte count;			// Number of MSGs in list.

} MSG_LIST;

/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static bool DeliverMsg(MSG_LIST *pList);

static void SendMsg(
	MSG_LIST *pList, byte dest, byte id, word wData);
static void AddTail(MSG_LIST *pList, MSG *pMsg);
MSG *RemoveHead(MSG_LIST *pList);
void UnlinkFromList(MSG_LIST *pList, MSG *pMsg);

static MSG *AllocateMsg(void);

/***************************************************************/

//		*** VARIABLES ***

static MSG msgPool[NUM_MSGS];	// Pool of messages.

static MSG_LIST freeList;		// List of free messages in pool.
static MSG_LIST msgList;		// Messages to be delivered.
static MSG_LIST eventList;		// Events to be delivered.


//static MSG *pCurrEvent;		// Pointer to event being processed
static byte currentTask;	// Task number of current task.

static unsigned long systemTime;	// System clock.  Counts mS up.

	// Task state pointers.

static void (*tasks[NUM_TASKS])(MSG *pEvent);

	// Timers

static TIMER timers[NUM_TIMERS];

/***************************************************************/

//		*** PROGRAM ***


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Fatal Error Detected

void Ex_FatalError(char *file, int line, byte data)
{
	cli();	// non-nested disable interrupts
	App_FatalError(file, line, data);
	while(1);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void Ex_SendEvent(byte dest, byte id)
{
	Ex_SendEventExt(dest, id, (word)0);
}

void Ex_SendEventExt(byte dest, byte id, word wData)
{
	SendMsg(&eventList, dest, id, wData);
}

void Ex_SendMsg(byte dest, byte id)
{
	Ex_SendMsgExt(dest, id, (word)0);
}

void Ex_SendMsgExt(byte dest, byte id, word wData)
{
	SendMsg(&msgList, dest, id, wData);
}

static void SendMsg(MSG_LIST *pList, byte dest, byte id, word wData)
{
	MSG *pMsg;
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	// Get a message from the list and fill it in:

	pMsg = AllocateMsg();

	pMsg->dest = dest;
	pMsg->state = MS_SENT;
	pMsg->id = id;
	pMsg->wData = wData;

	// Link message to end of selected MSG_LIST:

	AddTail(pList, pMsg);

	EI(ieStatus);		// Nested enable interrupts
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// The following routines add and remove messages from message lists.
// Interrupts are assumed to have been disabled by the caller.

// Add a message to the end of the list.

static void AddTail(MSG_LIST *pList, MSG *pMsg)
{
	pMsg->pPrev = pList->pLast;

	// If the 'last' pointer is null, the list is empty.

	if(pList->pLast != NULL)
		pList->pLast->pNext = pMsg;
	else
		pList->pFirst = pMsg;

	pList->pLast = pMsg;
	pMsg->pNext = NULL;
}

// Remove a message from the head of the list and return
// a pointer to it to the caller.

MSG *RemoveHead(MSG_LIST *pList)
{
	MSG *pMsg = pList->pFirst;

	if(pMsg != NULL)
		UnlinkFromList(pList, pMsg);

	return(pMsg);
}

// Unlink a message from anywhere in the list.

void UnlinkFromList(MSG_LIST *pList, MSG *pMsg)
{
	// If previous msg is NULL, this msg is at head of list.

	if(pMsg->pPrev == NULL)
		pList->pFirst = pMsg->pNext;
	else
		pMsg->pPrev->pNext = pMsg->pNext;

	// If next msg is NULL, this msg is at the end of the list.

	if(pMsg->pNext == NULL)
		pList->pLast = pMsg->pPrev;
	else
		pMsg->pNext->pPrev = pMsg->pPrev;
}


//	Allocate a message from the free message list.
// Interrupts assumed to have been disabled by caller.

static MSG *AllocateMsg(void)
{
	MSG *pMsg;

	if((pMsg = RemoveHead(&freeList)) == NULL)
		FATAL;

	return(pMsg);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Read the system clock:

unsigned long Ex_GetSystemTime(void)
{
	return(systemTime);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Start a polling timer.  This timer will call the
	// service address upon expriation, and will automatically
	// reload.  It counts in units of mS and may be set for up
	// to 65535 mS.

void Ex_StartPollTimer(
	byte timerNum, word duration, void (*pServiceFunc)(void))
{
	byte ieStatus;
	TIMER *pTmr;

	DI(ieStatus);		// Nested disable interrupts

	pTmr = &timers[timerNum];

	if(pTmr->state != TS_STOPPED)
		Ex_StopTimer(timerNum);

	pTmr->type = TT_POLL_TIMER;
	pTmr->state = TS_RUNNING;
	pTmr->prescaler = pTmr->resetPrescaler = 1;
	pTmr->timeCount = pTmr->resetTime = duration;
	pTmr->pServiceFunc = pServiceFunc;

	EI(ieStatus);		// Nested enable interrupts
}

	// Start an event timer.  This timer will send an event
	// upon expriation and will not reload.  It counts in units
	// indicated by the prescaler.

void Ex_StartEventTimer(
	byte timerNum, word prescaler, word duration, byte id)
{
	byte ieStatus;
	TIMER *pTmr;

	DI(ieStatus);		// Nested disable interrupts

	pTmr = &timers[timerNum];

	if(pTmr->state != TS_STOPPED)
		Ex_StopTimer(timerNum);

	pTmr->type = TT_EVENT_TIMER;
	pTmr->state = TS_RUNNING;
	pTmr->prescaler = pTmr->resetPrescaler = prescaler;
	pTmr->timeCount = pTmr->resetTime = duration;

	pTmr->event.dest = currentTask;
	pTmr->event.id = id;
	pTmr->event.state = MS_FREE;
	pTmr->event.allocType = MT_TIMER;

	EI(ieStatus);		// Nested enable interrupts
}


	// Stop the timer indicated by the timerNum parameter.
	// If the timer is an event timer, check to see if it
	// is timed-out and if so, delete (unlink) the event also.

void Ex_StopTimer(byte timerNum)
{
//	byte i;
	MSG *pMsg;
	TIMER *pTmr;

	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

	pTmr = &timers[timerNum];
	if(pTmr->type == TT_EVENT_TIMER && pTmr->state == TS_TIMED_OUT)
	{
		// If the event has not been processed, unlink it.

		pMsg = &pTmr->event;

		if(pMsg->state == MS_SENT)
		{
			UnlinkFromList(&eventList, pMsg);
			pMsg->state = MS_FREE;
		}
	}

	pTmr->state = TS_STOPPED;

	EI(ieStatus);		// Nested enable interrupts
}

bool Ex_TimerIsRunning(byte timerNum)
{
	return(timers[timerNum].state == TS_RUNNING);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set state machine execution address

void Ex_SetState(void (*pState)(MSG *pEvent))
{
	tasks[currentTask] = pState;
}


void Ex_InitTask(byte taskNum, void (*pInitAddr)(void))
{
	currentTask = taskNum;
	(*pInitAddr)();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Main:

//	This function is passed control by the C start-up
//	file.  It performs power-up initialization, then
//	goes into a loop in which it retreives an event
//	and distributes the event to the proper state
//	machine.

void main(void)
{
	byte i;

	App_InitIO();			// Initialize application I/Os

	// Initialize the message lists (pointers are zeroed by C init):

	for(i=0; i<NUM_MSGS; i++)
	{
		msgPool[i].allocType = MT_POOL;
		AddTail(&freeList, &msgPool[i]);
	}

	// systemTime = 0;				// Actually done by C init.

	// Timer 0 Initialization: This timer is used for general
	// purpose software timer generation.  It is initialized for
	// a 1 ms timebase.

	TCNT0 = TMR0_COUNT;	// Set initial count

	// Select 1/64 prescaler and start timer.

#ifdef __AVR_ATmega168__
	TCCR0B = TMR0_START;
#elif defined __AVR_ATmega8__
	TCCR0 = TMR0_START;
#endif

	// Init Timer Interrupts

#ifdef __AVR_ATmega168__
	TIMSK0 |= BV(TOIE0);
#elif defined __AVR_ATmega8__
	TIMSK |= BV(TOIE0);
#endif

	App_InitTasks();			// Initialize the application tasks

	sei();	// non-nested enable interrupts
Serial_TxChar('x');

	FOREVER
	{
		cli();	// non-nested disable interrupts

		// Check to see if there are any messages to deliver.
		// Messages from the "msgList" are higher priority,
		// messages from the "eventList" are lower priority.

		if(!DeliverMsg(&msgList))
			DeliverMsg(&eventList);

		sei();	// non-nested enable interrupts
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Deliver a message from one of the message queues.

static bool DeliverMsg(MSG_LIST *pList)
{
	MSG *pMsg;

	if((pMsg = RemoveHead(pList)) != NULL)
	{
		pMsg->state = MS_DELIVERED;
		currentTask = pMsg->dest;
Serial_TxChar(currentTask | 0x30);
		sei();	// non-nested enable interrupts
		(*tasks[currentTask])(pMsg);
		cli();	// non-nested disable interrupts
Serial_TxChar('x');

		pMsg->state = MS_FREE;
		if(pMsg->allocType == MT_POOL)
			AddTail(&freeList, pMsg);	// Deallocate message.

		return(TRUE);
	}

	else
		return(FALSE);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//	Timer 0 Overflow Handler

//	This interrupt generates general purpose time bases for
//	software generated timers.  Time bases of 1, 10, and
//	100 ms, as well as 1 and 10 seconds are generated.
//	Software timers may base their counting rates on any
//	of the available time bases.

SIGNAL(SIG_OVERFLOW0)
{
DB_Hi();

	byte i;
	TIMER *pTmr;

	TCNT0 += TMR0_COUNT;	// Refresh the timer

	systemTime++;			// Increment system time.

	// Enable interrupts during this interrupt.  This
	// allows higher priority interrupts such as the
	// audio generator to execute.

//	sei();	// non-nested enable interrupts

	for(i=0, pTmr = &timers[0]; i<NUM_TIMERS; i++, pTmr++)
	{
		if(pTmr->state == TS_RUNNING)
		{
//if(i==TM_MORSE_DIT)
//{
//	DB_PrintWord(pTmr->prescaler);
//	DB_PrintWord(pTmr->timeCount);
//	DB_PrintWord(pTmr->type);
//	DB_Mark(1);
//}

			if(--pTmr->prescaler == 0)
			{
				pTmr->prescaler = pTmr->resetPrescaler;
				if(--pTmr->timeCount == 0)
				{
					if(pTmr->type == TT_POLL_TIMER)
					{
//						cli();	// non-nested disable interrupts
						pTmr->timeCount = pTmr->resetTime;
						(*pTmr->pServiceFunc)();
//						sei();	// non-nested enable interrupts
					}

					else if(pTmr->type == TT_EVENT_TIMER)
					{
//						if(i==TM_MORSE_DIT)
//							DB_Mark(1);
//	DB_PrintWord(rate * timeCount);

//						cli();	// non-nested disable interrupts
						pTmr->state = TS_TIMED_OUT;
						pTmr->event.state = MS_SENT;
						AddTail(&eventList, &pTmr->event);
//						sei();	// non-nested enable interrupts
					}

					else
						FATAL;
				}
			}
		}
	}

DB_Low();

//	cli();	// non-nested disable interrupts
}

