/***************************************************************
*																					*
*	morse.c: Waveform Generator Audio Output							*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "bitio.h"
#include "morse.h"
#include "util.h"
#include "serial.h"
#include "parse.h"
#include "fgen.h"		// Waveform Generation

#ifdef DEBUG
#include "cdebug.h"
#endif

/***************************************************************/

//		*** MACROS ****

	//	Morse Cose Definitions

	// The prescaler stored in EEPROM is the number that
	// changes when we change the wpm rate.  The other
	// definitions determine the ratio in time between dits,
	// dahs, and spaces.

#define MCDIT_T			1	// Duration of DIT
#define MCDAH_T			3	// Duration of DAH
#define MCBIT_SPACE_T	1	// Space after DIT or DAH
#define MCCHAR_SPACE_T	3	// Space after char
#define MCWORD_SPACE_T	7	// Space after word


	// Times in mS.

#define TIME_PREMORSE	500	// Pre-Morse timer, approx. 0.5 sec
#define TIME_POSTMORSE	500	// Post-Morse timer, approx. 0.5 sec


/***************************************************************/

//		*** GENERAL DEFINITIONS ****


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void MorseXlate(void);
static void SendNextMorseBit(void);
static void SendMorseBit(byte time);
static void StartMDitTmr(byte timeCount);
static void StartMDlyTmr(word timeCount);
static void DoCommonStateProcessing(MSG *pEvent);

static void Transition_State_A(void);
static void State_A(MSG *pEvent);
static void Transition_State_B(void);
static void State_B(MSG *pEvent);
static void Transition_State_C(void);
static void State_C(MSG *pEvent);
static void Transition_State_E(void);
static void State_E(MSG *pEvent);
static void Transition_State_F(void);
static void State_F(MSG *pEvent);


/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** VARIABLES ***

static byte morseBuf[Z_WORDSTRING]; // Buf for outgoing msgs
static byte *pMorseStr;				// Points to current pos in buffer
static byte morseByte;				// Byte we are currently sending
static byte bitCount;				// Count of dits/dahs in curr char.
static word wMorseLevel;			// Volume of morse code tone.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Morse Code Table

//	Morse code is encoded as follows:
//	Bits 0-2: Length of Character, 1 to 5
//	Bits 3-7: Dits and Dahs, 1=dah, 0=dit

//	Exceptions:
//		Word space: 0x80 (since 0x00 is not allowed).
//		6 dit/dah codes:
//			Bits 3-7 contain the last 5 bits of the
//				character.
//			Bits 0-2 = 6 for characters with leading
//				dit, 7 for characters with
//				leading dah.

static prog_uchar MorseTbl[] =
{
	'A',0b00010010,	// 00 A: .-
	'B',0b00001100,	// 01 B: -...
	'C',0b00101100,	// 02 C: -.-.
	'D',0b00001011, 	// 03 D: -..
	'E',0b00000001,	// 04 E: .
	'F',0b00100100, 	// 05 F: ..-.
	'G',0b00011011,	// 06 G: .--
	'H',0b00000100,	// 07 H: ....
	'I',0b00000010,	// 08 I: ..
	'J',0b01110100,	// 09 J: .---
	'K',0b00101011,	// 10 K: -.-
	'L',0b00010100,	// 11 L: .-..
	'M',0b00011010,	// 12 M: --
	'N',0b00001010,	// 13 N: -.
	'O',0b00111011,	// 14 O: ---
	'P',0b00110100,	// 15 P: .--.
	'Q',0b01011100,	// 16 Q: --.-
	'R',0b00010011,	// 17 R: .-.
	'S',0b00000011,	// 18 S: ...
	'T',0b00001001,	// 19 T: -
	'U',0b00100011,	// 20 U: ..-
	'V',0b01000100, 	// 21 V: ...-
	'W',0b00110011,	// 22 W: .--
	'X',0b01001100, 	// 23 X: -..-
	'Y',0b01101100,	// 24 Y: -.--
	'Z',0b00011100, 	// 25 Z: --..

	'0',0b11111101,	// 26 0: -----
	'1',0b11110101,	// 27 1: .----
	'2',0b11100101,	// 28 2: ..---
	'3',0b11000101,	// 29 3: ...--
	'4',0b10000101,	// 30 4: ....-
	'5',0b00000101,	// 31 5: .....
	'6',0b00001101,	// 32 6: -....
	'7',0b00011101,	// 33 7: --...
	'8',0b00111101,	// 34 8: ---..
	'9',0b01111101,	// 35 9: ----.

	'/',0b01001101,	// 36 /: -..-.
	'-',0b10001101, 	// 37 BT: -...-

	// 6 dit/dah characters:

	'.',0b10101110,	// 38 .: .-.-.-
	'?',0b00110110,	// 39 ?: ..--..

	' ',0b10000000,	// 40 Space

	0x00, 0x00			// marks end of table
};

#define MAX_MORSE_CHAR	40		// Index must be <= this.

char Morse_GetCharFromTable(byte index)
{
	if(index > MAX_MORSE_CHAR)
		return(0x00);
	else
		return((char)LO_BYTE(PRG_RDW((word *)MorseTbl + index)));
}

//	This table encodes several rates for morse code output.

static prog_uchar morseRateTbl[] =
{
	120,		//  0: 10 wpm
	109,		//  1: 11 wpm
	100,		//  2: 12 wpm
	92,		//  3: 13 wpm
	86,		//  4: 14 wpm
	80,		//  5: 15 wpm
	75,		//  6: 16 wpm
	71,		//  7: 17 wpm
	67,		//  8: 18 wpm
	63,		//  9: 19 wpm
	60,		// 10: 20 wpm
	57,		// 11: 21 wpm
	55,		// 12: 22 wpm
	52,		// 13: 23 wpm
	50,		// 14: 24 wpm
	48			// 15: 25 wpm

				// Set MAX_MORSE_CODE_RATE and
				// MIN_MORSE_CODE_RATE in "h" file.
};


/*****************************************************************

								Global Functions

	The following functions provide the external interface (API)
	for the Morse Code State Machine.

*****************************************************************/

// Reset initialization of the Morse Code State Machine.  Called
// by the exec.c module upon reset.

void Morse_Init(void)
{
	Transition_State_F();		// Select state F to start.
}


/*****************************************************************

								Static Functions

	The following functions provide the internal functionality
	of the Morse Code State Machine and are not accessable by
	external modules.

*****************************************************************/

// The following functions start and stop the two timers used
// during morse code output.  These functions are called from
// within the morse code state machine.

static void StartMDitTmr(byte timeCount)
{
	byte index;
	byte rate;

	// Get rate from EEPROM and convert it into a table index.

	index = EEP_GetMorseRate() - MIN_MORSE_CODE_RATE;

	rate = PRG_RDB((byte *)morseRateTbl + index);

	Ex_StartEventTimer(TM_MORSE_DIT, TP_1MS,
		(rate * timeCount), EV_MORSE_DIT_TO);
}

#define StopMDitTmr()	Ex_StopTimer(TM_MORSE_DIT)


static void StartMDlyTmr(word timeCount)
{
	Ex_StartEventTimer(
		TM_MORSE_DLY, TP_1MS, timeCount, EV_MORSE_DLY_TO);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Translate Morse Characters in Buffer

//	This function scans the morse buffer and translates
//	all characters to their morse code bit encodings.

static void MorseXlate(void)
{
	byte c, c2;
	byte *pTable;
	char *pString = morseBuf;

	while(*pString != '\0')
	{
		pTable = (byte *)MorseTbl;	// Point at start of table

		c2 = *pString;
		c2 = (c2 >= 'a' && c2 <= 'z')? c2 & ~0x20 : c2;

		FOREVER
		{
			c = PRG_RDB(pTable++);
			if(c == c2 || c == ' ')
			{
				*pString = PRG_RDB(pTable++);
				break;	// FOREVER
			}
			else
				pTable++;
		}

		pString++;
	}

	pMorseStr = morseBuf;
	wMorseLevel = EEP_GetMorseLevel();
	StartMDlyTmr(TIME_PREMORSE);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// State F: Idle.  Wait until an event is received that starts
//	a morse activity.


static void Transition_State_F(void)
{
	// Select next state = F
	Ex_SetState(State_F);

#ifdef MORSE_DEBUG
	SerialTxChar('f');
#endif

}

static void State_F(MSG *pEvent)
{
	switch(pEvent->id)
	{
		DoCommonStateProcessing(pEvent);

		// Here, copy a string from flash and send it in Morse.

		case MSG_MORSE_FLASH:
			MovStrFSZ(morseBuf, (char *)pEvent->wData);
			MorseXlate();
			Transition_State_A();
			break;

		// Here, copy a string from SRAM and send it in Morse.

		case MSG_MORSE_SRAM:
			CopyStr(morseBuf, (char *)pEvent->wData);
			MorseXlate();
			Transition_State_A();
			break;

		default:
			break;

	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Morse State A

//	Get a character to send via morse.  Assume the string
//	has already been translated via morse_tbl.

static void Transition_State_A(void)
{
	// Select next state = A
	Ex_SetState(State_A);

#ifdef MORSE_DEBUG
	SerialTxChar('a');
#endif

}

static void State_A(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	// Wait for a timer timeout

	if(pEvent->id == EV_MORSE_DLY_TO || pEvent->id == EV_MORSE_DIT_TO)
	{
		if((morseByte = *pMorseStr++) != '\0')
		{
			if(morseByte == 0x80)	// Is it an inter-word space?
				StartMDitTmr(MCWORD_SPACE_T);

			// Here, not inter-word space

			else
			{
				bitCount = morseByte & 0x07;
				morseByte = (morseByte >> 3) & 0x1F;

				if(bitCount == 0x06)			// 6 bit char w/ dit?
				{
					SendMorseBit(MCDIT_T);	// Send initial dit
					bitCount = 5;				// 5 bits to go.
				}
				else if(bitCount == 0x07)	// 6 bit char w/ dah?
				{
					SendMorseBit(MCDAH_T);	// Send initial dah
					bitCount = 5;				// 5 bits to go.
				}

				else
				{
					bitCount--;
					SendNextMorseBit();
				}

				Transition_State_B();
			}
		}

		// Here we are all done sending.

		else
		{
			StopMDitTmr();
			StartMDlyTmr(TIME_POSTMORSE);
			Transition_State_E();
		}
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	MorseStateB

//	Here, delay for one inter-bit time (same as dit).

static void Transition_State_B(void)
{
	// Select next state = B
	Ex_SetState(State_B);

#ifdef MORSE_DEBUG
	SerialTxChar('b');
#endif

}

static void State_B(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	// Wait for a timer timeout

	if(pEvent->id == EV_MORSE_DIT_TO)
	{
		FGen_Stop(FGW_TONE);	// Stop the freq generator.

		if(bitCount-- != 0)	// Are there more bits to go?
		{
			StartMDitTmr(MCBIT_SPACE_T);
			Transition_State_C();
		}

		// Here, that character is complete so delay one inter-
		// character time, then do next character.

		else
		{
			StartMDitTmr(MCCHAR_SPACE_T);
			Transition_State_A();
		}
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	MorseStateC:

//	Here, send next dit or dah.  Goes back to state B
//	when done.


static void Transition_State_C(void)
{
	// Select next state = C
	Ex_SetState(State_C);

#ifdef MORSE_DEBUG
	SerialTxChar('c');
#endif

}

static void State_C(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	// Wait for a timer timeout

	if(pEvent->id == EV_MORSE_DIT_TO)
	{
		SendNextMorseBit();
		Transition_State_B();
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	MorseStateE:

//	Post Morse Delay.  This state delays some amount of time
//	after sending a morse string before signaling end of
//	morse output.


static void Transition_State_E(void)
{
	// Select next state = E
	Ex_SetState(State_E);

#ifdef MORSE_DEBUG
	SerialTxChar('e');
#endif

}


static void State_E(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	// Wait for a timer timeout

	if(pEvent->id == EV_MORSE_DLY_TO)
	{
		Ex_SendEvent(SM_AUDIO, EV_MORSE_CMPLT);
		Transition_State_F();
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Here, send either a dit or a dah.  This function sets
//	the morse_tmr so the state machine will execute again
//	when the timer times out.

void SendNextMorseBit(void)
{
	if(morseByte & 0x01)
	{
		SendMorseBit(MCDAH_T);		// Send a Dah
#ifdef MORSE_DEBUG
		SerialTxChar('_');
#endif
	}

	else
	{
		SendMorseBit(MCDIT_T);		// Send a Dit
#ifdef MORSE_DEBUG
		SerialTxChar('.');
#endif
	}

	morseByte = morseByte >> 1;
}

void SendMorseBit(byte time)
{
	StartMDitTmr(time);
	FGen_SetFreq(FGW_TONE, (EEP_GetMorseFreq()*10));
	FGen_SetGain(FGW_TONE, wMorseLevel);
	FGen_Start(FGW_TONE);
}


static void DoCommonStateProcessing(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// Reduce volume during talk-overs:

		case MSG_MORSE_WISPER:
			wMorseLevel = EEP_GetMorseLevel() / 2;
			break;

		// Go back to normal volume:

		case MSG_MORSE_SHOUT:
			wMorseLevel = EEP_GetMorseLevel();
			break;

		default:
			break;
	}
}


/****************************************************************
		End of Morse Code Module
****************************************************************/
