/***************************************************************
*																					*
*	parse.c: A system to aid in parsing command lines entered	*
*	via a terminal or serial port		 									*
*	Command Line Parsing Functions										*
*																					*
*	The following functions parse a command line contained		*
*	in a memory buffer.														*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************

	// COMMANDS

	A									// Toggle Rx Audio Enable
	B <freqHz>, <level%>			// Output A Tone. 600-1200 Hz, 0-100%
	C <wpmRate>, <freqHz>, <level%>	// Set CW Parameters
	D <freqHz>, <delayMS>, <level%>	// Set Courtesy Beep Parameters
	E <string>						// Set DTMF Password
	F <txEna>, <plDEna>, <rev>, <beep>, <DMute> // Set Flags
	G <tone>							// Generate CTCSS Tone
	H 									// Set CTCSS Decode Timeout
	I <string>						// Set CW ID String
	J									// Toggle PTT on/off
	K <mode>							// Set ID and Announcement Mode
	L 									// Unused
	M <timeMin>						// Set Announcement Period
	N <mode>							// Set CTCSS Encode Mode
	O 									// Unused
	P <tone>							// Set PL Tone Decoder
	Q 									// Unused
	R 									// Record Voice ID
	S									// Print status
	T <id>, <timeout>, <hang>	// Set ID, Timeout and Hang Timers
	U									// Unused
	V									// Unused
	W									// Play back Voice or CW ID.
	X 8958							// Set All Parameters to Default
	Y									// Unused
	Z									// Testing & Debug

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#include <limits.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "parse.h"
#include "serial.h"
#include "plsm.h"
#include "txsm.h"
#include "morse.h"
#include "dtmf.h"
#include "util.h"
#include "bitio.h"
#include "fgen.h"
#include "audio.h"

#ifdef CDEBUG
#include "cdebug.h"
#endif

/***************************************************************/

//		*** MACROS ****

/***************************************************************/

//		*** GENERAL DEFINITIONS ****

#define Z_CMD_BUF		(Z_WORDSTRING*2)	// Buffer length
#define Z_CW_ID			Z_WORDSTRING
	// Minimum and Maximum Allowed Parameter Values

#define MIN_IDTIME		60			// Minimum ID period in Seconds
#define MAX_IDTIME		1920		// Maximum ID period in Seconds

#define MIN_TXTIME		10			// Minimum Tx timeout in Seconds
#define MAX_TXTIME		600		// Maximum Tx timeout in Seconds

#define MIN_HANGTIME		200		// Minimum hang period in mS
#define MAX_HANGTIME		10000		// Maximum hang period in mS

#define MIN_CBDELAY		100		// Minimum courtesy beep delay in mS
#define MAX_CBDELAY		10000		// Maximum courtesy beep delay in mS

#define MIN_PL_TIMEOUT	20			// Minimum PL Detector Timeout, mS
#define MAX_PL_TIMEOUT	2000		// Maximum PL Detector Timeout, mS

#define MIN_ANN_PERIOD	1			// Minimum Announcement Period, min.
#define MAX_ANN_PERIOD	4320		// Maximum Announcement Period, min.


		// Default Tone Level for "B" Command

		// Level that gives 1000 Hz tone at same output voltage
		// level as is obtained with a 1000 Hz tone at 1VRMS input.

#	define DEFAULT_TONE_LEVEL	75

//		*** Parse Error Codes ***

#define PERR_OK			'0'		// No error.
#define PERR_CMDNF		'1'		// Command not found.
//								'2'		// Not Used.
#define PERR_EXTRA		'3'		// Extra chars at end of line.
#define PERR_VALUE		'4'		// Improper value.
#define PERR_ILLCHR		'5'		// Illegal characters in parm.
#define PERR_NOPARM		'6'		// Parameter not present.


	// Definition of Flags in EEPROM Channel Flag Byte

#define CF_DMUTE_ENA		0x01		// DTMF Mute Enable.
#define CF_BEEP_ENA		0x02		// Enable Courtesy Beep if TRUE
#define CF_PL_DECODE		0x04		// Master PL Decode Flag
#define CF_TX_ENA			0x08		// Enable Transmitter if TRUE.
//#define 					0x10		// Unused
#define CF_PL_REVBURST	0x20		// Reverse burst if TRUE
#define CF_HANG_ENA		0x40		// Set if hang time enabled.
//#define 					0x80		// Unused.

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***

	// PL Encoder/Decoder Channel Data:

struct TAG_PL_CHAN
{
	byte enabled;
	byte mode;
	byte decodeNum;
	byte encodeNum;
};

/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static char DoCWRateFreqCmd(void);
static char DoIDCmd(void);
static char DoStatusCmd(void);
static char DoTimerCmd(void);
static char DoPasswordCmd(void);
static char DoDefaultCmd(void);
static char DoPTTCmd(void);
static char DoRxAudioCmd(void);
static char DoToneCmd(void);

static char DoRecordVidCmd(void);
static char DoIDModeCmd(void);
static char DoSetAnnPeriodCmd(void);

static char DoPlayIDCmd(void);
static char DoFlagsCmd(void);
static char DoCourtesyBeepFreqCmd(void);
//static char DoSetPLTimeoutCmd(void);

static char DoPLToneCmd(void);
static char DoEncodeCTCSSModeCmd(void);
static char ParseFlag(byte flag);

static char DoPLGenerateCmd(void);

#ifdef DEBUG
static char DoDebugCmd(void);
#endif

static void InitCmdBuf(void);
static char GetNextChar(void);
static void AdvanceChar(void);
static byte Parse(void);
static char ParseWhite(void);
static bool IsTerminator(char c);
static void EEPROMWrite(void);
static void SetEEPDefaults(void);
static byte EEPChecksum(void);
static void ToggleIO(byte *pState, byte ioAddr);
static void DisplayToggleState(byte *pState);

static char StringParm(char *pString, byte length);
static char DecParm(word *pNum, word llim, word eqlim, word ulim);

/***************************************************************/

//		*** EEPORM VARIABLES ***

struct TAG_EEPROM
{
	byte chanFlags;		// Uses definitions in chan.h
	byte idMode;			// ID Mode

	struct TAG_PL_CHAN plChan[MAX_PL_TONES];

//	word plTimeout;		// Decoder timeout in mS.
	byte plEncodeMode;	// Encoder mode.  See 'h' file.
//	byte plTweek;			// Sets PL Parameters.

	char idString[Z_CW_ID];				// CW ID string
	char passWord[Z_DTMF_PASSWORD];	// Password string

	word morseFreq;		// Frequency of morse output
	byte morseRate;		// Index into morse rate table.
	byte morseLevel;		// CW Audio Level

	word cbFreq;			// Courtesy Beep Frequency
	word cbDelay;			// Courtesy Beep Delay
	byte cbLevel;			// Courtesy Beep Audio Level

	word hangTime;			// Length of hang time in mS
	word txTime;			// Length of Tx timeout in Sec
	word idTime;			// Length of ID timer in Sec.

	word vidDuration;		// Voice ID Duration in mS.
	word annPeriod;		// Announcement Period

	byte checksum;			// Checksum of EEPROM data
};


/***************************************************************/

//		*** SRAM VARIABLES ***

static char cmdBuf[Z_CMD_BUF];
static char *pBuf;		// Pointer into buffer being parsed
static char charCount;	// Number of characters parsed.

static struct TAG_EEPROM sramImage;	// SRAM image of EEPROM data

static bool ptt;			// TRUE if PTT Power is on.
static bool tone;			// TRUE if 1KHz tone is on.
static bool rxAudio;		// TRUE if Rx Audio is on.
static bool bGenTone;	// TRUE if we are generating PL.

static word frequency;			// Frequency of tone.
static word level;				// % amplitude of tone.

static byte flagBuf;			// Buffers flags during flag parsing.
static byte changedFlags;	// Indicates which flags have changed

/****************************************************************

		*** Program Memory ***

****************************************************************/

	// EEPROM Defaults

	// SRAM image of EEPROM data

static struct TAG_EEPROM __attribute__((progmem)) eepDefaults =
{
		// Channel flags
	(CF_TX_ENA | CF_BEEP_ENA | CF_DMUTE_ENA | CF_HANG_ENA),

	MODE_ID_CW,							// CW ID Mode

		// Channel PL tones
	{
		{0, MODE_EDPL_DECONLY, 10, 10},	// Chan 0
		{0, MODE_EDPL_ENCDEC, 12, 12},	// Chan 1
		{0, MODE_EDPL_ENCDEC, 14, 14},	// Chan 2
		{0, MODE_EDPL_ENCDEC, 16, 16}		// Chan 3
	},

//	1000,									// PL Decode Timeout in mS.

	MODE_ECTCSS_RUS,					// Encode Mode

//	0,										// PL Tweek

#ifdef DEBUG
	{'K','K','7','X','O','/','R','\0'},	// ID
#else
	{'N','O',' ','C','W','I','D','\0'},	// ID
#endif

	{'9','5','8','\0'},				// 958: Default password

	800,									// Morse: 800 Hz
	20,									// Morse: 20 wpm
	50,									// Audio Level, %

	700,									// Courtesy Beep: 700 Hz
	500,									// Pre-Courtesy-Beep Delay, mS.
	50,									// Courtesy Beep Level %

	1500,									// Hang time in mS (1.5 Sec)

#ifdef DEBUG_SHORT_TIMEOUTS
	30,									// Tx timeout in Sec
	60,									// CW ID timeout in Sec.
#else
	180,									// Tx timeout in Sec (3 Min)
	420,									// CW ID timeout in Sec (7 Min).
#endif

	0,										// Voice ID Duration in mS.

	360,									// Announcement period (1/4 day)

	0										// Checksum
};


/***************************************************************/

//		*** API ***

	// Initialize the Parser.

void Parse_Init(void)
{
	// vidPlay =
	ptt = tone = rxAudio = bGenTone = FALSE;
}


//	Get a line from the terminal.

//	This function calls SerialRxChar which waits in a loop
//	for the next character from the UART, thus hanging the
//	CPU.

void Parse_InitLexLine(void)
{
	charCount = 0;			// Clear char count
	pBuf = cmdBuf;
	Serial_TxFlashString((char *)PSTR("\r\n> "));
}

void Parse_LexLine(char c)
{
	byte parseErrors;

	if(c != '\r')
	{
		// Here, we received a backspace.
		if(c == '\b')
		{
			if(charCount > 0)
			{
				charCount--;
				pBuf--;
				Serial_TxFlashString((char *)PSTR("\b \b"));
			}
		}

		// Here, a space or printable character,

		else if((c >= '!' && c <= '~') || c == ' ')
		{
			if(charCount < Z_CMD_BUF-1)
			{
				Serial_TxChar(c);
				*pBuf++ = c;
				charCount++;
			}
		}

		else
			Serial_TxChar('\a');	// Ring the bell
	}

	// Here, we received a new line character.  End the
	// buffer and parse the line.

	else
	{
		*pBuf = '\0';

		if((parseErrors = Parse()) == PERR_OK)
			Serial_TxFlashString((char *)PSTR(" OK"));
		else
		{
			Serial_TxFlashString((char *)PSTR(" ERR "));
			Serial_TxChar(parseErrors);
		}

		Parse_InitLexLine();		// Init for next line.
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Get a characters from the command buffer

static void InitCmdBuf(void){pBuf = cmdBuf;}

static char GetNextChar(void) {return(*pBuf);}

static void AdvanceChar(void) {if(*pBuf != '\0') pBuf++;}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Parse a command buffer.

//	Input:	Pointer to null-terminated command buffer
//				Pointer to command table (byte address!)

//	Output:	TRUE if parsing errors, otherwise FALSE.

// Parser syntax error key:
//		Error 1: Command name not found.
//		Error 2: Command name must be followed by space or
//					end of line.
//		Error 3: Extra characters at end of command. (Too many
//					parameters for command).
//		Error 4: Parameters contained improper values for
//					the command, or all necessary parameters
//					were not entered.
//		Error 5: Parameter improperly formed.  (Alpha character
//					embedded in decimal parameter, etc.)

static byte Parse(void)
{
	char c;

	// Check to see if the line is empty.  If so, it is not
	// necessarilly an error.

	InitCmdBuf();

	if((c = GetNextChar()) != '\0')
	{
		AdvanceChar();		// Point at next char.

		// Here, we have a character on the command line.  Try to
		// match it in the table.

		switch(c)
		{
			// Toggle Rx Audio Enable

			case 'A':
			case 'a':
				return(DoRxAudioCmd());

			// Output a 1 KHz Tone
			case 'B':
			case 'b':
				return(DoToneCmd());

			// Set CW Rate and Frequency
			case 'C':
			case 'c':
				return(DoCWRateFreqCmd());

			// Set courtesy beep frequency
			case 'D':
			case 'd':
				return(DoCourtesyBeepFreqCmd());

			// Set DTMF Password
			case 'E':
			case 'e':
				return(DoPasswordCmd());

			// Change flags
			case 'F':
			case 'f':
				return(DoFlagsCmd());

			// Generate PL Tone

			case 'G':
			case 'g':
				return(DoPLGenerateCmd());

			// Set PL Timeout Timer

//			case 'H':
//			case 'h':
//				return(DoSetPLTimeoutCmd());

			// Set CW ID String
			case 'I':
			case 'i':
				return(DoIDCmd());

			// Toggle PTT
			case 'J':
			case 'j':
				return(DoPTTCmd());

			// Set ID Mode
			case 'K':
			case 'k':
				return(DoIDModeCmd());

			// Unused
//			case 'L':
//			case 'l':
//				return(());

			// Set Announcement Period
			case 'M':
			case 'm':
				return(DoSetAnnPeriodCmd());

			// Set CTCSS Encode Mode

			case 'N':
			case 'n':
				return(DoEncodeCTCSSModeCmd());

			// Set PL Tone to Decode:
			case 'P':
			case 'p':
				return(DoPLToneCmd());

			// Set PL Tone to Encode:

//			case 'Q':
//			case 'q':
//				return(DoEncodePLToneCmd());

			// Record Voice ID
			case 'R':
			case 'r':
				return(DoRecordVidCmd());

			// Display status to User
			case 'S':
			case 's':
				return(DoStatusCmd());

			// Set ID, Timeout and Tail Timers
			case 'T':
			case 't':
				return(DoTimerCmd());

			// Unused
//			case 'U':
//			case 'u':
//				return(());

			// Unused
//			case 'V':
//			case 'v':
//				return(());

			// Play Voice or CW ID
			case 'W':
			case 'w':
				return(DoPlayIDCmd());

			// Set All Parameters to Default
			case 'X':
			case 'x':
				return(DoDefaultCmd());

#ifdef DEBUG
			// Testing & Debugging
			case 'Z':
			case 'z':
				return(DoDebugCmd());
#endif

			default:
				return(PERR_CMDNF);		// Command Not Found
		}
	}

	return(PERR_OK);		// No errors.
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Parse whitespace from command buffer

static char ParseWhite(void)
{
	char c;

	while((c = GetNextChar()) != '\0')
	{
//DB_BYTE(c);

		if(c != ' ')
			break;
		AdvanceChar();		// Point at next char.
	}
	return(c);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Return TRUE if character is a terminator,
//	otherwise return FALSE.

static bool IsTerminator(char c)
{
	if(c == '\0' || c == ' ' || c == ',')
		return TRUE;
	else
		return FALSE;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Parse an unsigned decimal parameter into *pNum.  Maximum
//	number that can be entered here is 9999.

//	Input:	pNum: Buffer for parsed decimal number.
//				llim: Lower limit. Number must be greater than this.
//				eqlim: Equal limit.  Number is OK if equal to this.
//				ulim: Upper Limit.  Number must be less than this.

//	Output:	PERR_OK is returned if no errors, otherwise an
//				error code is returned.
//				Buffer pointer is returned pointing at terminator.

static char DecParm(word *pNum, word llim, word eqlim, word ulim)
{
	char parseError = PERR_OK;	// No error.
	word num = 0;
	char c;

	charCount = 0;
	ParseWhite();			// Get past white space.

	FOREVER
	{
		c = GetNextChar();

//DB_BYTE(c);

		if(IsTerminator(c))
		{
			if(charCount == 0)
				parseError = PERR_NOPARM;	// No parameter present.

			else if((num > llim && num < ulim) || num == eqlim)
				*pNum = num;

			else
				parseError = PERR_VALUE;	// Improper value.

			break;
		}

		if(c < '0' || c > '9')
		{
			parseError = PERR_ILLCHR;	// Illegal characters
			break;
		}

		c &= 0x0F;

		if((num >= 6553) && (c > 5))
		{
			parseError = PERR_VALUE;	// Illegal values.
			break;
		}

		// Multiply num by 10 and add c.

		num = (((num << 2) + num) << 1) + c;

		charCount++;
		AdvanceChar();		// Point at next char.
	}

	return(parseError);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Parse String Parameter.  Must be last parameter on a line
	// (So that " " and "," etc. can be part of the string)

	//	Buffer pointer is returned pointing at terminator.

static char StringParm(char *pString, byte length)
{
	char c;

	charCount = 0;
	ParseWhite();			// Get past white space.

	FOREVER
	{
		c = GetNextChar();

		if(c == '\0')
		{
			*pString = '\0';
			return(PERR_OK);
		}

		*pString = c;

		if(++charCount == length)
		{
			*pString = '\0';
			return(PERR_VALUE);
		}

		pString++;
		AdvanceChar();		// Point at next char.
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Parse a flag that can be either 0 or 1.

	//	Buffer pointer is returned pointing at terminator.

static char ParseFlag(byte flag)
{
	char parseError;
	word flagNum;

	changedFlags |= flag;

	if((parseError = DecParm(&flagNum, 0, 0, 2)) != PERR_OK)
		return(parseError);
	else
	{
		if(flagNum == 1)
			flagBuf |= flag;
		else
			flagBuf &= ~flag;
	}

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//				*** Commands ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Toggle Rx Audio Gate

	//	Format: A

static char DoRxAudioCmd(void)
{
	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	ToggleIO(&rxAudio, O_RXENA);

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// B: Output a Tone

	// B <freq>, <level> (both can default, see below)

static char DoToneCmd(void)
{
	char parseError;			// Parser error code.

	if(tone)
	{
		FGen_Stop(FGW_TONE);
		tone = FALSE;

		IO_Clear(O_PTT);
		ptt = FALSE;
	}

	else
	{
		parseError = DecParm(&frequency, 600, 600, 1201);

		if(parseError == PERR_NOPARM)
		{
			frequency = 1000;		// Default to 1000 Hz.
			level = DEFAULT_TONE_LEVEL;
		}

		else if(parseError == PERR_OK)
		{
			AdvanceChar();
			parseError = DecParm(&level, 0, 0, 101);

			if(parseError == PERR_NOPARM)
				level = DEFAULT_TONE_LEVEL;

			else
			{
				if(parseError != PERR_OK)
					return(parseError);
			}
		}

		else
			return(parseError);

		if(GetNextChar() != '\0')
			return(PERR_EXTRA);

		// Here, no parse errors.

		FGen_SetFreq(FGW_TONE, (frequency*10));
		FGen_SetGain(FGW_TONE, level);
		FGen_Start(FGW_TONE);
		tone = TRUE;


		IO_Clear(O_RXENA);
		rxAudio = FALSE;

		IO_Set(O_PTT);
		ptt = TRUE;
	}

	DisplayToggleState(&tone);

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set CW Rate and Frequency

	//	Format: C <rate>, <freq>, <level>
	// All parameters must be present.

static char DoCWRateFreqCmd(void)
{
	char parseError;
	word morseRate, morseFreq, morseLevel;

	// Get Parameters:

	if((parseError = DecParm(&morseRate, MIN_MORSE_CODE_RATE-1,
		MAX_MORSE_CODE_RATE, MAX_MORSE_CODE_RATE)) != PERR_OK)
		return(parseError);

	AdvanceChar();		// Get past terminator.

	if((parseError = DecParm(&morseFreq, FG_MIN_FREQ-1,
		FG_MAX_FREQ, FG_MAX_FREQ)) != PERR_OK)
		return(parseError);

	AdvanceChar();		// Get past terminator.

	if((parseError = DecParm(
		&morseLevel, 0, 0, 101)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.morseRate = morseRate;
	sramImage.morseFreq = morseFreq;
	sramImage.morseLevel = (byte)morseLevel;
	FGen_SetGain(FGW_TONE, sramImage.morseLevel);
	EEPROMWrite();

	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set Courtesy Beep Frequency and Pre-Courtesy-Beep Delay

	//	Format: D <freqHz> <delayMS> <level%>

static char DoCourtesyBeepFreqCmd(void)
{
	char parseError;
	word cbFreq, cbDelay, cbLevel;

	// Get Parameters:

	if((parseError = DecParm(&cbFreq, FG_MIN_FREQ-1,
		FG_MAX_FREQ, FG_MAX_FREQ)) != PERR_OK)
		return(parseError);

	AdvanceChar();		// Get past terminator.

	if((parseError = DecParm(&cbDelay, MIN_CBDELAY-1,
		MAX_CBDELAY, MAX_CBDELAY)) != PERR_OK)
		return(parseError);

	AdvanceChar();		// Get past terminator.

	if((parseError = DecParm(
		&cbLevel, 0, 0, 101)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.cbFreq = cbFreq;
	sramImage.cbDelay = cbDelay;
	sramImage.cbLevel = (byte)cbLevel;

	EEPROMWrite();

	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set DTMF Password String

	//	Format: E <string> \r

static char DoPasswordCmd(void)
{
	char parseError;
	char stringBuf[Z_DTMF_PASSWORD];	// String parameter
	byte i;

	if((parseError =
		StringParm(stringBuf, Z_DTMF_PASSWORD)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	// Check: string must be numeric.

	for(i=0; i<(Z_DTMF_PASSWORD-1); i++)
	{
		if(stringBuf[i] < '0' || stringBuf[i] > '9')
			return(PERR_ILLCHR);
	}

	if(stringBuf[Z_DTMF_PASSWORD-1] != '\0')
		return(PERR_ILLCHR);

	CopyStr(sramImage.passWord, stringBuf);

	EEPROMWrite();
	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set Flags: Transmit Enable, CTCSS Enable, Voice ID Enable,
	// and Courtesy Beep Enable.

	//	Format: F <beepEna>, <DTMFmute>, <hangEna>, <plDEna>,
	//					<plEEna>, <txEna>, <vid>

static char DoFlagsCmd(void)
{
	char parseError = PERR_NOPARM;
	byte flag;
	char c;

	flagBuf = changedFlags = 0;

	while((c = ParseWhite()) != '\0')
	{
		switch(c)
		{
			// Courtesy Beep Enable
			case 'B':
			case 'b':
				flag = CF_BEEP_ENA;
				break;

			// DTMF Mute Enable
			case 'D':
			case 'd':
				flag = CF_DMUTE_ENA;
				break;

			// Hang Time Enable
			case 'H':
			case 'h':
				flag = CF_HANG_ENA;
				break;

			// Master PL Tone Decode
			case 'P':
			case 'p':
				flag = CF_PL_DECODE;
				break;

			// PL Tone Encoder Reverse Burst
			case 'R':
			case 'r':
				flag = CF_PL_REVBURST;
				break;

			// Transmit Enable
			case 'T':
			case 't':
				flag = CF_TX_ENA;
				break;

			default:
				return(PERR_ILLCHR);

		} // switch

		AdvanceChar();

		if((parseError = ParseFlag(flag)) != PERR_OK)
			return(parseError);

		AdvanceChar();

	} // while

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	// Set Parameters in EEPROM:

	if(parseError == PERR_OK)
	{
		sramImage.chanFlags &= ~changedFlags;
		sramImage.chanFlags |= flagBuf;
		EEPROMWrite();
	}

	return(parseError);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// G: Generate a PL Tone (Toggle on and off)

	// Format: G <tone code>

static char DoPLGenerateCmd(void)
{
	char parseError;			// Parser error code.
	word toneCode;

	if(bGenTone)
	{
		FGen_DisableCTCSSGeneration();
		bGenTone = FALSE;
		IO_Clear(O_PTT);
		ptt = FALSE;
	}

	else
	{
		if((parseError = DecParm(
			&toneCode, 0, 0, MAX_PL_DECODE+1)) == PERR_NOPARM)
			toneCode = 10;

		else
		{
			if(parseError != PERR_OK)
				return(parseError);

			if(GetNextChar() != '\0')
				return(PERR_EXTRA);
		}

		bGenTone = TRUE;
		ptt = TRUE;
		IO_Set(O_PTT);
		FGen_EnableCTCSSGeneration(toneCode);
	}

	DisplayToggleState(&bGenTone);

	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set PL Tone Decoder Timeout

	//	Format: H <time>

	// Units of <time> is mS.
/*
static char DoSetPLTimeoutCmd(void)
{
	char parseError;
	word time;

	// Get Parameters:

	if((parseError = DecParm(&time, MIN_PL_TIMEOUT,
		MIN_PL_TIMEOUT, MAX_PL_TIMEOUT+1)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.plTimeout = time;
	EEPROMWrite();

	return(PERR_OK);
};

*/
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Get CW ID String

	//	Format: I <string> \r

static char DoIDCmd(void)
{
	char parseError;
	char stringBuf[Z_CW_ID];	// String parameter

	if((parseError = StringParm(stringBuf, Z_CW_ID)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	CopyStr(sramImage.idString, stringBuf);

	EEPROMWrite();
	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// J: Toggle PTT

static char DoPTTCmd(void)
{
	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	ToggleIO(&ptt, O_PTT);

	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// K: Get ID Mode

	//	Format: K <mode> \r

static char DoIDModeCmd(void)
{
	char parseError;
	word mode;
	byte result;

	// Accept just about anything.  (Decoded and verified below).

	if((parseError = DecParm(
		&mode, MODE_ID_CW, MODE_ID_CW, MODE_ID_MANUAL_ANN+1)) != PERR_OK)
		result = parseError;

	else if(GetNextChar() != '\0')
		result = PERR_EXTRA;

	else
	{
		sramImage.idMode = (byte)mode;
		EEPROMWrite();
		if(mode == MODE_ID_TIMED_ANN)
			Audio_StartAnnouncementTimer();	// Start announcement timer.
		else
			Audio_StopAnnouncementTimer();

		result = PERR_OK;
	}

	return(result);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set Announcement Time

	//	Format: M <timeMin>

	// Units of <timeMin> is minutes.


static char DoSetAnnPeriodCmd(void)
{
	char parseError;
	word time;

	// Get Parameters:

	if((parseError = DecParm(&time, MIN_ANN_PERIOD,
		MIN_ANN_PERIOD, MAX_ANN_PERIOD+1)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.annPeriod = time;
	EEPROMWrite();

	Audio_StartAnnouncementTimer();	// Start announcement timer.

	return(PERR_OK);
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set CTCSS Encode Mode

	//	Format: N <mode>


static char DoEncodeCTCSSModeCmd(void)
{
	char parseError;
	word mode;

	if((parseError = DecParm(&mode, MODE_ECTCSS_RUS,
		MODE_ECTCSS_RUS, MODE_ECTCSS_PTT+1)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.plEncodeMode = (byte)mode;
	EEPROMWrite();

	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	Set PL Tone Encode/Decode Parameters

	//	Format:	P <chan>,<ena>,<mode>,<tone1>,<tone2>

	//		<chan> is the encode/decode channel number
	//		<ena> is 0 to disable or 1 to enable the channel,
	//		<mode> is the encode/decode mode (see below)
	//		<tone1> and <tone2> are optional codes identifying
	//			specific CTCSS frequencies.  A table containing
	//			these codes is contained in plsm.c

	//		If <mode> is encode-only or decode-only, only one
	//			tone code is specified.  If <mode> is encode-decode,
	//			both tone codes must be specified.  In this case
	//			the first code corresponds to the decode frequency
	//			and the second code corresponds to the encode
	//			frequency.

	//		It is possible to only specify <chan> and <ena> if the
	//			user wishes to simply enable or disable the channel.

	//		The <mode> parameter may only be set to encode-only on
	//			channel 0.

static char DoPLToneCmd(void)
{
	char parseError;
	word chan, ena, mode;
	word plNum1, plNum2;

	// Get Channel:

	if((parseError = DecParm(
		&chan, 0, 0, MAX_PL_TONES+1)) != PERR_OK)
		return(parseError);

	// Get Enable Flag:

	AdvanceChar();		// Get past terminator.

	if((parseError = DecParm(&ena, 0, 0, 2)) != PERR_OK)
		return(parseError);

	// Get Mode if channel 0 selected:

	if(chan == 0)
	{
		AdvanceChar();		// Get past terminator.

		if((parseError = DecParm(
			&mode, 0, 0, MAX_MODE_EDPL+1)) == PERR_NOPARM)
		{
			// If no mode parm then enable or disable channel 0.

			sramImage.plChan[0].enabled = ena;
			EEPROMWrite();
			PL_SetToneDecodeThresholds();
			return(PERR_OK);
		}

		// Here, check to see if mode parm was invalid.

		else if(parseError != PERR_OK)
			return(parseError);

		// Here we have a valid mode parameter for channel 0.
		// Get tone codes:

		else
		{
			AdvanceChar();		// Get past terminator.
			sramImage.plChan[0].mode = mode;
			sramImage.plChan[0].enabled = ena;

			if((parseError = DecParm(
				&plNum1, 0, 0, MAX_PL_DECODE+1)) != PERR_OK)
				return(parseError);

			else if(mode == MODE_EDPL_ENCONLY)
				sramImage.plChan[0].encodeNum = plNum1;

			else if(mode == MODE_EDPL_DECONLY)
				sramImage.plChan[0].decodeNum = plNum1;

			// Here, must be encode/decode.  Get second tone parm.

			else
			{
				AdvanceChar();		// Get past terminator.

				if((parseError = DecParm(
					&plNum2, 0, 0, MAX_PL_DECODE+1)) != PERR_OK)
					return(parseError);

				sramImage.plChan[0].decodeNum = plNum1;
				sramImage.plChan[0].encodeNum = plNum2;
			}

			EEPROMWrite();
			PL_SetToneDecodeThresholds();
			return(PERR_OK);
		}
	}

	// Channels 1-3:

	else
	{
		// Channels 1 thru 3 are encode/decode.

		AdvanceChar();		// Get past terminator.

		parseError = DecParm(
					&plNum1, 0, 0, MAX_PL_DECODE+1);

		if(parseError != PERR_NOPARM)
		{
			if(parseError != PERR_OK)
				return(parseError);

			else
			{
				AdvanceChar();		// Get past terminator.

				if((parseError = DecParm(
					&plNum2, 0, 0, MAX_PL_DECODE+1)) != PERR_OK)
					return(parseError);

				sramImage.plChan[chan].decodeNum = plNum1;
				sramImage.plChan[chan].encodeNum = plNum2;
				sramImage.plChan[chan].mode = MODE_EDPL_ENCDEC;

			}
		}

		sramImage.plChan[chan].enabled = ena;
		EEPROMWrite();
		PL_SetToneDecodeThresholds();
		return(PERR_OK);

	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Enable Recording of Voice ID

	//	Format: R

static char DoRecordVidCmd(void)
{
	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	Ex_SendMsg(SM_RX, MSG_VID_ENA);
	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Status Command

static char DoStatusCmd(void)
{
	byte i;

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	Serial_CRLF();
	Serial_CRLF();

	// Do "C" status:

	Serial_TxFlashString((char *)PSTR("C "));
	Serial_TxDecWord(sramImage.morseRate);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord(sramImage.morseFreq);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord(sramImage.morseLevel);
	Serial_CRLF();


	// Do "D" Status:

	Serial_TxFlashString((char *)PSTR("D "));
	Serial_TxDecWord((word)sramImage.cbFreq);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord((word)sramImage.cbDelay);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord((word)sramImage.cbLevel);
	Serial_CRLF();

	// Do "E" status:

	Serial_TxFlashString((char *)PSTR("E "));
	Serial_TxString(sramImage.passWord);
	Serial_CRLF();

	// Do "F" status:
	wdt_reset();		// Refresh watchdog timer.

	Serial_TxFlashString((char *)PSTR("F"));
	Serial_TxFlashString((char *)PSTR(" B"));
	Serial_TxBool(sramImage.chanFlags & CF_BEEP_ENA);
	Serial_TxFlashString((char *)PSTR(", D"));
	Serial_TxBool(sramImage.chanFlags & CF_DMUTE_ENA);
	Serial_TxFlashString((char *)PSTR(", H"));
	Serial_TxBool(sramImage.chanFlags & CF_HANG_ENA);
	Serial_TxFlashString((char *)PSTR(", P"));
	Serial_TxBool(sramImage.chanFlags & CF_PL_DECODE);

	Serial_TxFlashString((char *)PSTR(", R"));
	Serial_TxBool(sramImage.chanFlags & CF_PL_REVBURST);

	Serial_TxFlashString((char *)PSTR(", T"));
	Serial_TxBool(sramImage.chanFlags & CF_TX_ENA);

	Serial_CRLF();

	// Do "H" status:

//	Serial_TxFlashString((char *)PSTR("H "));
//	Serial_TxDecWord(sramImage.plTimeout);
//	Serial_CRLF();

	// Do "I" status:

	Serial_TxFlashString((char *)PSTR("I "));
	Serial_TxString(sramImage.idString);
	Serial_CRLF();

	// Do "K" Status:

	Serial_TxFlashString((char *)PSTR("K "));
	Serial_TxChar(sramImage.idMode | '0');
	Serial_CRLF();

	// Do "M" status:

	Serial_TxFlashString((char *)PSTR("M "));
	Serial_TxDecWord(sramImage.annPeriod);
	Serial_CRLF();

	// Do "N" status:

	Serial_TxFlashString((char *)PSTR("N "));
	Serial_TxDecWord(sramImage.plEncodeMode);
	Serial_CRLF();

	// Do "P" status:

	for(i=0; i<MAX_PL_TONES; i++)
	{
		Serial_TxFlashString((char *)PSTR("P "));
		Serial_TxDecWord(i);
		Serial_TxFlashString((char *)PSTR(", "));
		Serial_TxDecWord(sramImage.plChan[i].enabled);
		Serial_TxFlashString((char *)PSTR(", "));

		if(i == 0)
		{
			Serial_TxDecWord(sramImage.plChan[i].mode);
			Serial_TxFlashString((char *)PSTR(", "));
		}

		Serial_TxDecWord(sramImage.plChan[i].decodeNum);
		Serial_TxFlashString((char *)PSTR(", "));
		Serial_TxDecWord(sramImage.plChan[i].encodeNum);
		Serial_CRLF();
	}

	// Do "T" status:

	Serial_TxFlashString((char *)PSTR("T "));
	Serial_TxDecWord(sramImage.idTime);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord(sramImage.txTime);
	Serial_TxFlashString((char *)PSTR(", "));
	Serial_TxDecWord(sramImage.hangTime);
	Serial_CRLF();

	Serial_CRLF();

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set ID, Timeout and Tail Timers

	// Format: T <id>, <timeout>, <hang> \r
	// All parameters must be present.

static char DoTimerCmd(void)
{
	char parseError;
	word idTime, txTime, hangTime;

	// Get Parameters:

	if((parseError = DecParm(&idTime,
		MIN_IDTIME-1, 0, MAX_IDTIME+1)) != PERR_OK)
		return(parseError);

	AdvanceChar();

	if((parseError = DecParm(&txTime,
		MIN_TXTIME, MIN_TXTIME, MAX_TXTIME+1)) != PERR_OK)
		return(parseError);

	AdvanceChar();

	if((parseError = DecParm(&hangTime,
		MIN_HANGTIME, MIN_HANGTIME, MAX_HANGTIME+1)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	sramImage.idTime = idTime;
	sramImage.txTime = txTime;
	sramImage.hangTime = hangTime;
	EEPROMWrite();

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Play Back CW or Voice ID

	//	Format: W

static char DoPlayIDCmd(void)
{
	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	Ex_SendMsg(SM_TX, MSG_TX_IDCMD);
	return(PERR_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Copy EEPROM defaults from SRAM into EEPROM

	// Format: X <code> \r

static char DoDefaultCmd(void)
{
	char parseError;
	word codeNum;

	// Get Parameters:

	if((parseError = DecParm(
		&codeNum, 8957, 8958, 8959)) != PERR_OK)
		return(parseError);

	if(GetNextChar() != '\0')
		return(PERR_EXTRA);

	SetEEPDefaults();


	FGen_SetGain(FGW_TONE, sramImage.morseLevel);

	PL_SetToneDecodeThresholds();

	return(PERR_OK);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Z: Testing & Debugging

#ifdef DEBUG
static char SendDTMF(void);

static char DoDebugCmd(void)
{
	char c;

	while((c = GetNextChar()) != '\0')
	{
		AdvanceChar();		// Point at next char.

		// Here, we have a character on the command line.  Try to
		// match it in the table.

		switch(c)
		{
			// Do DTMF Codes:

			case 'C':
			case 'c':
				if((c = SendDTMF()) != PERR_OK) return(c);
				else return(PERR_OK);

			// Do DTMF Codes (Bracketed by Unsquelch/Squelch):

			case 'D':
			case 'd':
				Ex_SendEvent(SM_RX, EV_UNSQUELCHED);
				if((c = SendDTMF()) != PERR_OK) return(c);
				Ex_SendEvent(SM_RX, EV_SQUELCHED);
				return(PERR_OK);

			// Unused.

//			case 'I':
//			case 'i':
//				break;

			// Hang and wait for watchdog to reset.

			case 'R':
			case 'r':
				FOREVER ;

			// Go Squelched:

			case 'S':
			case 's':
				Ex_SendEvent(SM_RX, EV_SQUELCHED);
				break;

			// Go Unsquelched:

			case 'U':
			case 'u':
				Ex_SendEvent(SM_RX, EV_UNSQUELCHED);
				break;

			default:
				return(PERR_CMDNF);
		}
	}

	return(PERR_OK);
}

static char SendDTMF(void)
{
	char c;

	while((c = GetNextChar()) != '\0')
	{
		AdvanceChar();

		// Here, try to match char to DTMF code.

		switch(c)
		{
			case '0':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0A);
				break;

			case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8':
			case '9':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, c & 0x0F);
				break;

			case 'A':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0D);
				break;

			case 'B':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0E);
				break;

			case 'C':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0F);
				break;

			case 'D':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x00);
				break;

			case '*':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0B);
				break;

			case '#':
				Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, 0x0C);
				break;

			default:
				return(PERR_ILLCHR);
		}
	}
	return(PERR_OK);
}

#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//				*** Support Functions ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Display the state of a toggle on the monitor.

static void DisplayToggleState(bool *pState)
{
	if(*pState)
		Serial_TxFlashString((char *)PSTR(" ON"));
	else
		Serial_TxFlashString((char *)PSTR(" OFF"));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// If an IO is set, clear it.  Otherwise, set it.

static void ToggleIO(byte *pState, byte ioAddr)
{
	if(*pState)
	{
		IO_Clear(ioAddr);
		*pState = FALSE;
	}
	else
	{
		IO_Set(ioAddr);
		*pState = TRUE;
	}

	DisplayToggleState(pState);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//				*** EEPROM Functions ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


static void SetEEPDefaults(void)
{
	wdt_reset();		// Refresh watchdog timer.
	memcpy_P(&sramImage, &eepDefaults, sizeof(struct TAG_EEPROM));
	EEPROMWrite();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void EEPROMWrite(void)
{
	byte *pSramImage;

	word i;

	sramImage.checksum = 0;
	sramImage.checksum = -EEPChecksum();

	pSramImage = (byte *)&sramImage;

	for(i=0; i<sizeof(struct TAG_EEPROM); i++)
	{
		if(eeprom_read_byte((byte *)i) != *pSramImage)
		{
			wdt_reset();		// Refresh watchdog timer.
			eeprom_write_byte((byte *)i, *pSramImage);
		}

		pSramImage++;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Read the EEPROM data and compute the checksum

void EEP_Init(void)
{
	eeprom_read_block(&sramImage, 0, sizeof(struct TAG_EEPROM));

	if(EEPChecksum() != 0)
	{
		Serial_TxFlashString(
			(char *)PSTR("EEPROM Checksum Bad...Setting defaults."));
		SetEEPDefaults();
	}

	else
		Serial_TxFlashString(
			(char *)PSTR("EEPROM Checksum OK."));
}


	// Calculate checksum of EEPROM image.

static byte EEPChecksum(void)
{
	byte i, sum;
	byte *pMem = (byte *)&sramImage;

	sum = 0;

	for(i=0; i<sizeof(sramImage); i++)
		sum += *pMem++;

	return(sum);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//				Functions used to Get EEPROM variables
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Get user configurable flags associated with channel

byte EEP_GetFlags(void)
{
	return(sramImage.chanFlags);
}

	// Get a pointer to the CW ID String.

char *EEP_GetCWIDString(void)
{
	return(sramImage.idString);
}

char *EEP_GetDTMFPassword(void)
{
	return(sramImage.passWord);
}

	// Return the CW wpm rate to the caller.

byte EEP_GetMorseRate(void)
{
	return(sramImage.morseRate);
}

	// Return the CW tone frequency to the caller.

word EEP_GetMorseFreq(void)
{
	return(sramImage.morseFreq);
}

	// Return the CW tone amplitude to the caller.

byte EEP_GetMorseLevel(void)
{
	return(sramImage.morseLevel);
}

word EEP_GetCourtesyBeepFreq(void)
{
	return(sramImage.cbFreq);
}

word EEP_GetCBDelay(void)
{
	return(sramImage.cbDelay);
}

byte EEP_GetCourtesyBeepLevel(void)
{
	return(sramImage.cbLevel);
}

word EEP_GetHangTime(void)
{
	return(sramImage.hangTime);
}

word EEP_GetTxTimeout(void)
{
	return(sramImage.txTime);
}

word EEP_GetIDTimeout(void)
{
	return(sramImage.idTime);
}

	// Return TRUE if Courtesy Beep is Enabled.

bool EEP_CourtesyBeepEnabled(void)
{
	return((sramImage.chanFlags & CF_BEEP_ENA) != 0);
}

	// PL Tone Channel Functions:

	// Get PL Tone enable flag

bool EEP_GetChanPLEnable(byte chan)
{
	if(sramImage.plChan[chan].enabled == 0)
		return(FALSE);
	else
		return(TRUE);
}

	// Get PL Tone enable flag

bool EEP_GetMasterPLEnable(void)
{
	return((sramImage.chanFlags & CF_PL_DECODE) != 0);
}

	// Get PL Channel Mode Byte.

byte EEP_GetPLMMode(byte chan)
{
	return(sramImage.plChan[chan].mode);
}

	// Get PL Tone decode frequency.

byte EEP_GetPLDecodeNum(byte chan)
{
	return(sramImage.plChan[chan].decodeNum);
}

	// Get PL Tone encode frequency.

byte EEP_GetPLEncodeNum(byte chan)
{
	return(sramImage.plChan[chan].encodeNum);
}

	// Get CTCSS decode timeout.

//word EEP_GetPLDecodeTimeout(void)
//{
//	return(sramImage.plTimeout);
//}

	// Flags

bool EEP_HangTimeIsEnabled(void)
{
	return((sramImage.chanFlags & CF_HANG_ENA) != 0);
}

bool EEP_DTMFMuteIsEnabled(void)
{
	return((sramImage.chanFlags & CF_DMUTE_ENA) != 0);
}

	// Return CTCSS Encode Mode.

byte EEP_CTCSSEncodeMode(void)
{
	return(sramImage.plEncodeMode);
}

//byte EEP_GetPLTweek(void)
//{
//	return(sramImage.plTweek);
//}

bool EEP_RevBurstIsEnabled(void)
{
	return((sramImage.chanFlags & CF_PL_REVBURST) != 0);
}


bool EEP_TxIsEnabled(void)
{
	return((sramImage.chanFlags & CF_TX_ENA) != 0);
}


	// Return TRUE if the ID mode is CW, and
	// Return FALSE if the ID mode is Voice.

bool EEP_IDModeIsCW(void)
{
	if(sramImage.idMode == MODE_ID_VOICE)
		return(FALSE);
	else
		return(TRUE);
}

	// Return TRUE if announcement mode is enabled,
	// FALSE otherwise.

bool EEP_AnnouncementEnabled(void)
{
	if((sramImage.idMode == MODE_ID_CW)
		|| (sramImage.idMode == MODE_ID_VOICE))
		return(FALSE);
	else
		return(TRUE);
}


	// Return TRUE if Timed Announcement Mode is enabled,
	// FALSE otherwise.

bool EEP_TimedAnnouncementMode(void)
{
	if(sramImage.idMode == MODE_ID_TIMED_ANN)
		return(TRUE);
	else
		return(FALSE);
}

	// Return TRUE if Manual Announcement Mode is enabled,
	// FALSE otherwise.

bool EEP_ManualAnnouncement(void)
{
	if(sramImage.idMode == MODE_ID_MANUAL_ANN)
		return(TRUE);
	else
		return(FALSE);
}


// Let the ID'er read the duration of the voice
// ID file.

word EEP_GetVoiceIDDuration(void)
{
	return(sramImage.vidDuration);
}

// Let the ID'er read the duration of the voice
// ID file.

word EEP_GetAnnouncementPeriod(void)
{
	return(sramImage.annPeriod);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//				Functions used to Set EEPROM variables
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *



void EEP_SetCTCSSEncodeMode(byte mode)
{
	sramImage.plEncodeMode = mode;
	EEPROMWrite();
}

	// Copy CW ID string into EEPROM.

void EEP_SetCWIDString(char *pString)
{
	byte index = 0;
	char c;

	while((c = *pString++) != '\0')
		sramImage.idString[index++] = c;

	sramImage.idString[index] = '\0';

	EEPROMWrite();
}

void EEP_SetIDMode(byte mode)
{
	if(mode == MODE_ID_TIMED_ANN)
		Audio_StartAnnouncementTimer();
	else
		Audio_StopAnnouncementTimer();

	sramImage.idMode = mode;
	EEPROMWrite();
}

// Store the duration of the voice ID in EEProm.

void EEP_SetVoiceIDDuration(word duration)
{
	sramImage.vidDuration = duration;
	EEPROMWrite();
}


//void EEP_SetPLTweek(byte tweek)
//{
//	sramImage.plTweek = tweek;
//	EEPROMWrite();
//}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//	Flags:

void EEP_SetReverseBurstState(bool state)
{
	if(state)
		sramImage.chanFlags |= CF_PL_REVBURST;
	else
		sramImage.chanFlags &= ~CF_PL_REVBURST;
	EEPROMWrite();
}

	// Set CTCSS decode enable state.

void EEP_SetMasterPLDecodeState(bool state)
{
	if(state)
		sramImage.chanFlags |= CF_PL_DECODE;
	else
		sramImage.chanFlags &= ~CF_PL_DECODE;
	EEPROMWrite();
}

void EEP_SetChanPLDecodeState(byte chan, bool state)
{
	sramImage.plChan[chan].enabled = state;

	EEPROMWrite();
}

	// Set Courtesy Beep enable state.

void EEP_SetCourtesyBeepState(bool state)
{
	if(state)
		sramImage.chanFlags |= CF_BEEP_ENA;
	else
		sramImage.chanFlags &= ~CF_BEEP_ENA;
	EEPROMWrite();
}

	// Set DTMF Mute state.

void EEP_SetDTMFMuteState(bool state)
{
	if(state)
		sramImage.chanFlags |= CF_DMUTE_ENA;
	else
		sramImage.chanFlags &= ~CF_DMUTE_ENA;
	EEPROMWrite();
}

void EEP_SetTxEnableState(bool state)
{
	if(state)
		sramImage.chanFlags |= CF_TX_ENA;
	else
		sramImage.chanFlags &= ~CF_TX_ENA;
	EEPROMWrite();
}

