/***************************************************************
*																					*
*	serial.c: Functions for RS-232 Serial I/O							*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "serial.h"
#include "bitio.h"
#include "parse.h"
#include "cdebug.h"

/***************************************************************/

//		*** MACROS ****


/***************************************************************/

//		*** GENERAL DEFINITIONS ****


	//	Baud Rate Definitions
	// UBRR constant for 19.2K baud (2x mode for 1.0 PCB,
	// 1x mode for 2.0 PCB.)

#	define BAUD19K2	((((SYSCLK/(8*19.2))+0.5)/2)-1)	// 1x Mode

/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****


/***************************************************************/

//		*** DATA STRUCTURE Definitions ***

/***************************************************************/

//		*** VARIABLES ***


/***************************************************************/

//		*** PROGRAM ***


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Enable RS-232 Serial I/O

void Serial_Init(void)
{
//	byte ieStatus;
//	DI(ieStatus);		// Nested disable interrupts

	UBRR0H = HI_BYTE(BAUD19K2);
	UBRR0L = LO_BYTE(BAUD19K2);

	// Set the 2x Bit in the Baud Rate Generator for
	// 9600 or 19.2K Baud.  Clear it for 1200 Baud.

	UCSR0A &= ~BV(U2X0);

	 // Enable Rx and Tx, and enable Rx interrupts.

	UCSR0B = (BV(RXEN0) | BV(TXEN0) | BV(RXCIE0));

	// Set frame format: 8data, 2stop bit
	UCSR0C = (BV(USBS0) | BV(UCSZ00) | BV(UCSZ01));

//	EI(ieStatus);		// Nested enable interrupts

	Ex_SetState(Serial_SM);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Polled Serial I/O:  Stay in a spin loop until the port is
// ready, then send or receive a single character.

char Serial_RxChar(void)
{
	byte c;

	UCSR0B &= ~BV(RXCIE0);		// Disable receiver interrupts
	loop_until_bit_is_set(UCSR0A,RXC0);
	c = UDR0;
	UCSR0B |= BV(RXCIE0);		// Enable receiver interrupts

	return((char)c);
}

void Serial_TxChar(byte c)
{
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0 = c;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Additional Polled I/O Functions:


void Serial_TxBool(byte dat)
{
	Serial_TxChar(dat == 0? '0' : '1');
}


	//	This function converts a byte into two hex-ascii characters
	//	and outputs them to the UART.

void Serial_TxHexChar(byte dat)
{
	byte temp;
	temp = dat >> 4;
	temp = Serial_HexConv(temp);
	Serial_TxChar(temp);
	temp = Serial_HexConv(dat);
	Serial_TxChar(temp);
}


	// This function turns a word into it's decimal equivilent and
	// outputs it to the UART.

void Serial_TxDecWord(word data)
{
	char buffer[5];
	char *pC = &buffer[4];
	byte i;

	// Unpack the binary value into a decimal one.

	for(i=0, pC = &buffer[4]; i<5; i++, pC--)
	{
		*pC = (data % 10) | 0x30;
		data = data / 10;
	}

	// Send the decimal value to the serial port.
	// Supress leading zeros.

	for(i=0, pC = buffer; i<4; i++, pC++)
	{
		if(*pC != '0')
			break;
	}

	while(i<5)
	{
		Serial_TxChar(*pC++);
		i++;
	}
}

void Serial_TxString(char *pStr)
{
	char c;

	while((c = *pStr++) != '\0')
		Serial_TxChar(c);
}


void Serial_TxFlashString(char *pFlash)
{
	char c;

	while((c = PRG_RDB(pFlash++)) != '\0')
		Serial_TxChar(c);
}


void Serial_CRLF(void)
{
	Serial_TxFlashString((char *)PSTR("\r\n"));
}

//	This function converts a nibble to it's hex-ascii character.
// The nibble is contained in the low order 4 bits of 'nib'.

byte Serial_HexConv(byte nib)
{
	nib &= 0x0F;	// get low nibble only
	nib |= 0x30;	// set to ascii col 3
	if (nib >= 0x3A)
		nib += 0x07;	// move code to 0x41 to 0x46
	return nib;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// The UART receiver interrupt interface.

SIGNAL(SIG_USART_RECV)
{
	char c;

	c = UDR0;
	Ex_SendEventExt(SM_SIO, EV_SIO_RXRDY, (word)c);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Serial I/O State Machine:

void Serial_SM(MSG *pEvent)
{
	if(pEvent->id != EV_SIO_RXRDY)
		FATAL;
	else
	{
		Parse_LexLine((byte)pEvent->wData);	// Send character to line lexer
	}
}

