/***************************************************************
*																					*
*	txsm.c: Transmitter State Machine									*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "bitio.h"
#include "txsm.h"
#include "morse.h"
#include "serial.h"
#include "parse.h"
#include "audio.h"
#include "plsm.h"


#ifdef DEBUG
#include "cdebug.h"
#endif

		// undef define
#ifdef TX_DEBUG
#undef TX_DEBUG_LC_STATE_FLAGS	// Show state flags on monitor
#define TX_DEBUG_UC_STATE_FLAGS	// Show state flags on monitor
#undef TX_DEBUG_EV_DISP				// Show event display
#endif

/***************************************************************/

//		*** MACROS ****

#if defined TX_DEBUG_UC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y))
#elif defined TX_DEBUG_LC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y)|0x20)
#else
#define SET_STATE(X,Y)	Ex_SetState(X)
#endif



/***************************************************************/

//		*** GENERAL DEFINITIONS ****

#define DisablePTT()	IO_Clear(O_PTT)
#define EnablePTT()	IO_Set(O_PTT)


#define T_REVERSE_BURST		160	// Reverse burst for 160 ms.
//#define T_REVERSE_BURST		300	// Beta test


/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void IDTmrStart(void);
static void TxTimeout(void);

static void Transition_State_A(void);
static void State_A(MSG *pEvent);
static void Transition_State_B(void);
static void State_B(MSG *pEvent);
static void Transition_State_C(void);
static void Transition_State_C1(void);
static void State_C(MSG *pEvent);
static void Transition_State_D(void);
static void State_D(MSG *pEvent);

static void Transition_State_E(void);
static void State_E(MSG *pEvent);

//static bool EncodeOnRUS(void);
//static bool EncodeOnPTT(void);
//static void EnablePLEncoder(void);

//static void Transition_State_F(void);
//static void State_F(MSG *pEvent);
static void Transition_State_G(void);
static void State_G(MSG *pEvent);
static void Transition_State_H(void);
static void State_H(MSG *pEvent);

/***************************************************************/

//		*** VARIABLES ***

static bool endRepeatFlag;	// Used during timeout to detect end.
static bool txTimeoutFlag;	// TRUE if transmitter timed out.
static bool closingID;		// Flag set if need to do closing ID.
static bool openingID;		// Flag set if we can do opening ID.
static bool revBurstComplete;	// Reverse burst complete if true.

/***************************************************************/

//		*** Transmit State Machine API ***

void Tx_Init(void)
{
	endRepeatFlag = TRUE;
	revBurstComplete = TRUE;
	closingID = FALSE;
	openingID = TRUE;
	Transition_State_A();		// Select state A
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Start the ID Timer.

static void IDTmrStart(void)
{
	if(EEP_GetIDTimeout() != 0)
		Ex_StartEventTimer(
			TM_TX_ID, TP_SEC, EEP_GetIDTimeout(), EV_TX_IDTIME);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Start the TX (Alligator) Timer.

static void TXTmrStart(void)
{
	Ex_StartEventTimer(
		TM_TX_TIMEOUT, TP_SEC, EEP_GetTxTimeout(), EV_TX_TIMEOUT);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Process events and messages common to all states

static void DoCommonStateProcessing(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// Shut off tone encoder and flag a reverse burst
		// timeout

		case EV_TX_REVBURST:
			PL_DisableEncoder();
			revBurstComplete = TRUE;
			break;

		// Here, user dropped carrier.

		case MSG_TX_ENDRPT:
			endRepeatFlag = TRUE;
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Traansmitter State A:  Wait for receiver to go active
//	and tell transmitter to start repeating.

static void Transition_State_A(void)
{
	DisablePTT();

	PL_DisableEncoder();

	txTimeoutFlag = FALSE;
	SET_STATE(State_A, 'A');
}

static void State_A(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		case MSG_TX_REPEAT:
			if(EEP_TxIsEnabled())
			{
				// Enable the transmitters.

				if(openingID && (EEP_GetIDTimeout() != 0))
				{
					IDTmrStart();
					Audio_Schedule(AU_DO_ID);	// Schedule opening ID.
				}

				openingID=FALSE;			// Do not do opening ID again.
				closingID = TRUE;			// Do closing ID when idle.

				Transition_State_B();	// Repeating.
			}
			break;

		// If it becomes time to do an ID while idle, get it
		// started.

		case EV_TX_IDTIME:
			if(EEP_TxIsEnabled())
			{
				if(closingID)
				{
					IDTmrStart();					// Restart the timer.
					Audio_Schedule(AU_DO_ID);	// Schedule final ID.
					Transition_State_C1();		// Wait til complete.
					closingID = FALSE;
				}
				else
					openingID=TRUE;				// OK to do opening ID.
			}
			break;

		// Here, play an announcement.

		case EV_TX_ANNTO:
			if(EEP_TxIsEnabled())
			{
				// Enable the transmitters.

				if(openingID && (EEP_GetIDTimeout() != 0))
				{
					IDTmrStart();
					Audio_Schedule(AU_DO_ID);	// Schedule opening ID.
				}

				openingID=FALSE;			// Do not do opening ID again.
				closingID = TRUE;			// Do closing ID when idle.
				Transition_State_C1();	// Start announcement.
			}
			break;

		// Here, do an ID as requested by serial or DTMF.

		case MSG_TX_IDCMD:
			if(EEP_TxIsEnabled())
			{
				Audio_Schedule(AU_DO_ID);	// Schedule ID.
				Transition_State_C1();		// Wait til complete.
			}
			break;

		// Here, turn transmitter on.

		case MSG_TX_ON:
 			if(EEP_GetIDTimeout() != 0)
			{
				closingID = TRUE;				// Do ID when idle if needed.
				IDTmrStart();
				Audio_Schedule(AU_DO_ID);	// Schedule ID.
			}
			Transition_State_C1();		// Start Audio
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Transmitter State B:  We are repeating now.

static void Transition_State_B(void)
{
	EnablePTT();
	TXTmrStart();
	endRepeatFlag = FALSE;

	if(PL_EncodeOnPTTisEnabled() || PL_EncodeOnRUSisEnabled())
		PL_EnableEncoder();

	revBurstComplete = FALSE;

	// Select next state = B
	SET_STATE(State_B, 'B');
}

static void State_B(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// End repeating function.
		// Do courtesy beep.

		case MSG_TX_ENDRPT:
			if(EEP_CourtesyBeepEnabled())
				Audio_Schedule(AU_DO_CB);

			if(PL_EncodeOnRUSisEnabled())
			{
				// If reverse burst, must change output
				// phase and delay PTT for 160 ms.

				if(PL_RevBurstIsEnabled())
				{
					Ex_StartEventTimer(TM_REVERSE_BURST, TP_1MS,
						T_REVERSE_BURST, EV_TX_REVBURST);
					PL_ReverseBurst();
				}

				// Here, no reverse burst so just shut off
				// encoder.
				else
					PL_DisableEncoder();
			}

			// In any case (If 'no encode' or 'encode mode = PTT',
			// go to C.

			Transition_State_C();
			break;

		// Here, been transmitting too long.

		case EV_TX_TIMEOUT:
			TxTimeout();
			break;

		// Here, ID timer timed out.  Schedule an ID.

		case EV_TX_IDTIME:
			Audio_Schedule(AU_DO_ID);
			IDTmrStart();
			break;

		// Here, we are receiving DTMF codes.  Turn off the
		// transmitters and wait for DTMF complete.  Don't
		// turn off the tx timer because we can use it to
		// time out the DTMF command.

		case MSG_DTMF_INPROG:
			Transition_State_H();
			break;

		default:
			break;
	}
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


// Transmitter State C: Wait for audio response to complete.

static void Transition_State_C(void)
{
	EnablePTT();
	Audio_Start();
	Ex_StopTimer(TM_TX_TIMEOUT);

	// Select next state = C
	SET_STATE(State_C, 'C');
}

	// If we have CTCSS Encoding on PTT enabled, we need to
	// start encoding for announcements and IDs here.

static void Transition_State_C1(void)
{

	if(PL_EncodeOnPTTisEnabled())
	{
		PL_EnableEncoder();
		revBurstComplete = FALSE;
	}

	else
	{
		if(PL_EncodeOnRUSisEnabled())
			revBurstComplete = TRUE;
	}

	Transition_State_C();
}

static void State_C(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Here, If we haven't timed out, go back to repeating.
		case MSG_TX_REPEAT:
			if(!txTimeoutFlag && EEP_TxIsEnabled())
				Transition_State_B();	// Repeating.
			break;

		// Here, the audio state machine has finished it's last
		// task.

		case EV_AUD_CMPLT:

			// If the transmitter timed out, must wait for user
			// to drop the carrier.

			if(txTimeoutFlag)
			{
				if(endRepeatFlag)
					Transition_State_A();	// If carrier dropped, go idle.
				else
					Transition_State_G();	// else wait for carrier to drop.
			}
			else
			{
				// If Reverse Burst is enabled we have to delay
				// PTT for 160 mS.

				if(PL_EncodeOnPTTisEnabled())
				{
					// If reverse burst, must change output
					// phase and delay PTT for 160 ms.
					if(PL_RevBurstIsEnabled())
					{
//	Serial_TxChar('1');

						Ex_StartEventTimer(TM_REVERSE_BURST, TP_1MS,
							T_REVERSE_BURST, EV_TX_REVBURST);
						PL_ReverseBurst();

						if(EEP_HangTimeIsEnabled())
							Transition_State_D();		// Do hang time.
						else
							Transition_State_E();
					}

					// Here, no reverse burst.

					else
					{
						PL_DisableEncoder();

						if(EEP_HangTimeIsEnabled())
							Transition_State_D();		// Do hang time.
						else
							Transition_State_A();
					}
				}

				// Here, we should have started reverse burst
				// already.  See if it's complete.

				else if(PL_EncodeOnRUSisEnabled())
				{
					// If reverse burst is enabled and is not
					// complete, wait it out.
					if(PL_RevBurstIsEnabled())
					{
						if(EEP_HangTimeIsEnabled())
							Transition_State_D();		// Do hang time.
						else
							Transition_State_E();		// Wait for rev burst cp.

					}
					else
					{
						if(EEP_HangTimeIsEnabled())
							Transition_State_D();		// Do hang time.
						else
							Transition_State_A();		// Else drop PTT.
					}
				}

				// Here, no tone encoding selected.
				else
				{
					if(EEP_HangTimeIsEnabled())
						Transition_State_D();		// Do hang time.
					else
						Transition_State_A();		// Else drop PTT.
				}
			}
			break;

		// Here, ID timer timed out.

		case EV_TX_IDTIME:
			Audio_Schedule(AU_DO_ID);
			IDTmrStart();
			break;

		default:
			break;
	}

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Transmitter State D: Wait out the hang time and then go
// idle.  If the user attempts to transmit while hang time is
// counting, go to state B to resume repeating.

static void Transition_State_D(void)
{
	Ex_StartEventTimer(
		TM_TX_HANG, TP_1MS, EEP_GetHangTime(), EV_TX_HANG);

	// Select next state = D
	SET_STATE(State_D, 'D');
}

static void State_D(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Come here if hang timer times out.

		case EV_TX_HANG:
			// If reverse burst is enabled and is not
			// complete, wait it out.

//			if(PL_EncodeOnPTTisEnabled() && PL_RevBurstIsEnabled())
//			{
//				Serial_TxChar('1');

				// If reverse burst, must change output
				// phase and delay PTT for 160 ms.
//				Ex_StartEventTimer(TM_REVERSE_BURST, TP_1MS,
//					T_REVERSE_BURST, EV_TX_REVBURST);
//				PL_ReverseBurst();
//				Transition_State_E();
//			}

//			else
			if(PL_RevBurstIsEnabled())
			{
//				Serial_TxChar('2');

				if(revBurstComplete)
					Transition_State_A();
				else
					Transition_State_E();
			}
			else
			{
//				Serial_TxChar('3');
				Transition_State_A();
			}
			break;

		case MSG_TX_REPEAT:
			if(EEP_TxIsEnabled())
			{
				Ex_StopTimer(TM_TX_HANG);	// Stop hang timer.
				Transition_State_B();		// Repeating.
			}
			break;

		// Here, ID timer timed out.  Schedule an ID.

		case EV_TX_IDTIME:
			if(EEP_TxIsEnabled())
			{
				Audio_Schedule(AU_DO_ID);
				IDTmrStart();

				if(PL_EncodeOnPTTisEnabled())
				{
					PL_EnableEncoder();
					revBurstComplete = FALSE;
				}

				Transition_State_C();
			}
			break;

		// Here, play an announcement.

		case EV_TX_ANNTO:
			if(EEP_TxIsEnabled())
			{
				// Should we turn on the tone encoder?

				if(PL_EncodeOnPTTisEnabled())
				{
					PL_EnableEncoder();
					revBurstComplete = FALSE;
				}

				Transition_State_C();		// Wait til complete.
			}
			break;

		default:
			break;
	}

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Transmitter State E: Wait for reverse burst to complete.

static void Transition_State_E(void)
{
	// Select next state = E
	SET_STATE(State_E, 'E');
}

static void State_E(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Here, If we haven't timed out, go back to repeating.
		case MSG_TX_REPEAT:
			if(EEP_TxIsEnabled())
			{
				Ex_StopTimer(TM_REVERSE_BURST);
				Transition_State_B();	// Repeating.
			}
			break;

		// Here, ID timer timed out.

		case EV_TX_IDTIME:
			if(EEP_TxIsEnabled())
			{
				Ex_StopTimer(TM_REVERSE_BURST);
				Audio_Schedule(AU_DO_ID);
				IDTmrStart();

				if(PL_EncodeOnPTTisEnabled())
				{
					PL_EnableEncoder();
					revBurstComplete = FALSE;
				}

				Transition_State_C();
			}
			break;

		case EV_TX_REVBURST:
			Transition_State_A();	// Go Idle.
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Transmitter State G: After Tx timeout message sent, wait for
//	user to unkey.

static void Transition_State_G(void)
{
	DisablePTT();

	PL_DisableEncoder();

	// Select next state = G
	SET_STATE(State_G, 'G');
}

static void State_G(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		case MSG_TX_ENDRPT:
			Transition_State_A();
			break;

		default:
			break;
	}
}

static void TxTimeout(void)
{
	Ex_SendEvent(SM_RX, EV_RX_TXTO);	// Tell Rx.
	Audio_Schedule(AU_DO_TO);

 	if(EEP_GetIDTimeout() != 0)
	{
		Audio_Schedule(AU_DO_ID);
		Ex_StopTimer(TM_TX_ID);
	}

	Ex_StopTimer(TM_REVERSE_BURST);
	txTimeoutFlag = TRUE;
	openingID=TRUE;				// OK to do opening ID.
	Transition_State_C();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Transmitter State H: Here, DTMF codes were detected during
// repeating state.  Wait here until DTMF commands are complete
//	then give a response.

static void Transition_State_H(void)
{
	if(EEP_DTMFMuteIsEnabled())
	{
		PL_DisableEncoder();
		DisablePTT();
	}

	// Select next state = H
	SET_STATE(State_H, 'H');
}

static void State_H(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// End DTMF reception.  Play status OK or ERR.

		case MSG_TX_ENDRPT:
			Transition_State_C1();	// Do audio functions.
			break;

		// Here, ID timer timed out.

		case EV_TX_IDTIME:
			Audio_Schedule(AU_DO_ID);
			IDTmrStart();
			break;

		// Here, been receiving DTMF too long.

		case EV_TX_TIMEOUT:
			TxTimeout();
			break;

		// Here, do an ID and turn the transmitter off.

		case MSG_TX_OFF:
			EEP_SetTxEnableState(FALSE);
			Audio_Schedule(AU_DO_ID);
			Ex_StopTimer(TM_TX_ID);
			closingID = FALSE;
			openingID = TRUE;
			break;

		// W issued from keyboard or DTMF equivilent

		case MSG_TX_IDCMD:
			Audio_Schedule(AU_DO_ID);
			Ex_StopTimer(TM_TX_ID);
			closingID = FALSE;
			openingID = TRUE;
			break;

		default:
			break;
	}
}

