/***************************************************************
*																					*
*	app.c: Application Initialization									*
*																					*
*	Copyright 2006 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "rxsm.h"
#include "txsm.h"
#include "bitio.h"
#include "serial.h"
#include "morse.h"
#include "plsm.h"
#include "dtmf.h"
#include "parse.h"
#include "fgen.h"
#include "audio.h"

#ifdef DEBUG
#include "cdebug.h"
#endif



/***************************************************************/

//		*** MACROS ****


/***************************************************************/

//		*** GENERAL DEFINITIONS ****


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void WDog_Task(MSG *pEvent);
static void WDog_Init(void);

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** CONSTANTS ***

static char __attribute__((progmem)) clrScrn[] =
	{0x1b,'[','2','J','\0'};		// Clear screen.


/***************************************************************/

//		*** VARIABLES ***


/***************************************************************/

//		*** PROGRAM ***


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	Initialize Application I/O

void App_InitIO(void)
{
	wdt_reset();		// Refresh watchdog timer.
	wdt_enable(WDTO_120MS);

	// Initialize I/Os

	IO_Init();

	// Set up A/D

	ADMUX = 0x62;		// Select ADC2, Left Adjust, AVCC ref.
//	ADCSRB = 0x05;		// Trigger on OC1B.
//	ADCSRB = 0x00;		// Free-Run Mode

	DIDR0 = 0x40;		// Disable digital I/O on pin ADC2

	// Start Timer 1 and A/D Conversion.  Enable A/D Interrupts.

	ADCSRA = 0x87;		// Enable, Start, Auto Trig, Clear IRQ
							// Enable IRQ, 128 Prescaler.

	// Initialize the Power Reduction Register (ATMega168):

#ifdef __AVR_ATmega168__
	PRR = BV(PRTWI) | BV(PRSPI) /* | BV(PRADC) */;
#endif

}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Initialize the Application State Machines

void App_InitTasks(void)
{

	// Sign on to the RS-232 port.

	Parse_Init();
	Ex_InitTask(SM_SIO, Serial_Init);	// Init serial I/O

	Serial_TxFlashString((char *)clrScrn);

	wdt_reset();		// Refresh watchdog timer.

#ifdef __AVR_ATmega168__
	Serial_TxHexChar(MCUSR);
#elif defined __AVR_ATmega8__
	Serial_TxHexChar(MCUCSR);
#endif

	Serial_TxChar(' ');
	Serial_TxFlashString((char *)PSTR(SIGNON1));
	Serial_TxFlashString((char *)PSTR(SIGNON2));

	EEP_Init();	// Here, read EEPROM and set initial values.

	Serial_CRLF();

	Parse_InitLexLine();	// Init parser.

	FGen_Init();	// Init waveform generator

	wdt_reset();		// Refresh watchdog timer.

	Ex_InitTask(SM_MORSE, Morse_Init);	// Init morse code generator
	Ex_InitTask(SM_PL, PL_Init);			// Init PL tone detection
	Ex_InitTask(SM_DTMF, DTMF_Init);		// Init DTMF detection
	Ex_InitTask(SM_RX, Rx_Init);			// Init squelch system
	Ex_InitTask(SM_TX, Tx_Init);			// Init transmitter control
	Ex_InitTask(SM_AUDIO, Audio_Init);	// Init Audio control
	Ex_InitTask(SM_WDOG, WDog_Init);		// Init Watchdog Task
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#define T_WDOG			50		// Watchdog times out in 128 ms
									// so refresh it in less time.

static void WDog_Init(void)
{
	Ex_StartEventTimer(
		TM_WDOG, TP_1MS, T_WDOG, EV_WDOG_REFRESH);
	Ex_SetState(WDog_Task);
}


static void WDog_Task(MSG *pEvent)
{
	switch(pEvent->id)
	{
		case EV_WDOG_REFRESH:
			wdt_reset();		// Refresh watchdog timer.
			Ex_StartEventTimer(
				TM_WDOG, TP_1MS, T_WDOG, EV_WDOG_REFRESH);
			break;

		default:
			FATAL;
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Do App Dependant Fatal Error Stuff

void App_FatalError(char *file, int line, byte data)
{
	IO_Clear(O_PTT);	// Disable transmitter.
	Serial_TxFlashString((char *)PSTR("\r\nFatal Error: "));
	Serial_TxString(file);
	Serial_TxFlashString((char *)PSTR(" line "));
	Serial_TxHexChar(HI_BYTE(line));
	Serial_TxHexChar(LO_BYTE(line));
	Serial_TxFlashString((char *)PSTR(", data "));
	Serial_TxHexChar(data);
	Serial_CRLF();
}
