/***************************************************************
*																					*
*	audio.c: Audio Control State Machine								*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*	Modified for single channel use in PSE-508 controller.		*
*																					*
*	Copyright 2006 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "bitio.h"
#include "morse.h"
#include "parse.h"
#include "fgen.h"
#include "audio.h"
#include "serial.h"

#ifdef DEBUG
#include "cdebug.h"
#endif


	// undef define

#ifdef AU_DEBUG
#undef AU_DEBUG_UC_STATE_FLAGS	// Show UC state flags on monitor
#define AU_DEBUG_LC_STATE_FLAGS	// Show LC state flags on monitor
#undef AU_DEBUG_EV_DISP			// Show event messages
#endif

#if defined AU_DEBUG_UC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y))
#elif defined AU_DEBUG_LC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y)|0x20)
#else
#define SET_STATE(X,Y)	Ex_SetState(X)
#endif

/***************************************************************/

//		*** MACROS ****


/***************************************************************/

//		*** GENERAL DEFINITIONS ****

#define T_VID_MUTE		10			// Mute VID to eliminate click
#define T_CB_DURATION	100		// Courtesy beep duration
#define T_ANNOUNCE_POLL	TP_MIN	// Announcement polling period

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** CONSTANTS IN FLASH ***

static prog_uchar DTMFErrString[] = {'E','R','R',0x00};

static prog_uchar DTMFOKSting[] = {'O','K',0x00,0x00};

static prog_uchar TOString[] = {'T','O',0x00,0x00};

static prog_uchar NVString[] = {
									'N', 'O', ' ', 'V', 'I', 'D', 0x00};

//static prog_uchar CourtesyBeep[] = {'T',0x00};

/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void Transition_State_A(void);
static void State_A(MSG *pEvent);
static void Transition_State_B(void);
static void State_B(MSG *pEvent);

#ifdef VOICE_ID
static void Transition_State_C(void);
static void State_C(MSG *pEvent);
static void Transition_State_D(void);
static void State_D(MSG *pEvent);
#endif // VOICE_ID

static void Transition_State_E(void);
static void State_E(MSG *pEvent);
static void Transition_State_F(void);
static void State_F(MSG *pEvent);
//static void Transition_State_G(void);
//static void State_G(MSG *pEvent);

static void DoNextTask(void);
static void DoMorseFlashMsg(byte flag, char *pString);
static void AudioComplete(void);
static void DoCommonStateProcessing(MSG *pEvent);
static void PollAnnounceISR(void);

/***************************************************************/

//		*** VARIABLES ***

static byte tasks;		// Scheduled audio tasks to perform.
static bool bUnsq;		// Unsquelched - generate PL if true.
static word announcementPeriod;

/*****************************************************************

								Global Functions

	The following functions provide the external interface (API)
	for the Receiver State Machine.

*****************************************************************/

static void PollAnnounceISR(void)
{
	if(--announcementPeriod == 0)
	{
		announcementPeriod = EEP_GetAnnouncementPeriod();
		Audio_Schedule(AU_DO_ANNO);
		Ex_SendMsg(SM_TX, EV_TX_ANNTO);	// Notify Tx Task
	}
}


void Audio_Init(void)
{
	IO_Set(O_VID_MUTE);			// Mute voice ID
	bUnsq = FALSE;

	if(EEP_TimedAnnouncementMode())
		Audio_StartAnnouncementTimer();

	Transition_State_A();		// Select state A
}


void Audio_StartAnnouncementTimer(void)
{
	byte ieStatus;

	Ex_StopTimer(TM_ANNOUNCE_POLL);
	Ex_StartPollTimer(
		TM_ANNOUNCE_POLL, T_ANNOUNCE_POLL, PollAnnounceISR);

	DI(ieStatus);		// Nested disable interrupts
	announcementPeriod = EEP_GetAnnouncementPeriod();
	EI(ieStatus);		// Nested enable interrupts
}


void Audio_StopAnnouncementTimer(void)
{
	Ex_StopTimer(TM_ANNOUNCE_POLL);
	Audio_Cancel(AU_DO_ANNO);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Set or clear an activity flag.  The activity flags are used
//	to keep track of the things the morse state machine has to
//	do when activated.


void Audio_Schedule(byte flag)
{
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

	tasks |= flag;

	EI(ieStatus);		// Nested enable interrupts
}

void Audio_Cancel(byte flag)
{
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

	tasks &= ~flag;

	EI(ieStatus);		// Nested enable interrupts
}

void Audio_Start(void)
{
	Ex_SendMsg(SM_AUDIO, MSG_AUD_START);
}

/*****************************************************************

								Static Functions

	The following functions provide the internal functionality
	of the Audio State Machine and are not accessable by
	external modules.

*****************************************************************/

// Audio State A: Idle, wait for a command to start.

static void Transition_State_A(void)
{
	// Select next state = A
	SET_STATE(State_A, 'A');
}

static void State_A(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		case MSG_AUD_START:
			DoNextTask();	// Selects new state also.
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Audio State B: Wait for Morse to Complete.

	// If squelch opens, reduce tone gain by 50%.

static void Transition_State_B(void)
{
	// Wisper the morse ID if the squelch is open.

	if(bUnsq)
		Ex_SendMsg(SM_MORSE, MSG_MORSE_WISPER);

	// Select next state = B
	SET_STATE(State_B, 'B');
}

static void State_B(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		case EV_MORSE_CMPLT:
			DoNextTask();
			break;

		// Reduce tone volume during talk-over:

		case MSG_AUD_UNSQ:
			Ex_SendMsg(SM_MORSE, MSG_MORSE_WISPER);
			break;

		// Increase tone volume if talk-over finished:

		case MSG_AUD_SQ:
			Ex_SendMsg(SM_MORSE, MSG_MORSE_SHOUT);
			break;


		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Audio State C: Wait here for VID mute time.

#ifdef VOICE_ID

static void Transition_State_C(void)
{
	// Start the Voice ID Duration Timer.  Playback stops when
	// this timer times out.  The duration is stored in EEPROM.

	Ex_StartEventTimer(
		TM_AUDIO_TMR1, TP_1MS, T_VID_MUTE, EV_AUD_TMR1);

	IO_Set(O_VIDPLAY);		// Start voice ID output

	// Select next state = C
	SET_STATE(State_C, 'C');
}

static void State_C(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Come here when mute timer times out.

		case EV_AUD_TMR1:
			Transition_State_D();		// Wait for VID to complete.
			break;

		// Here, someone is trying to talk over the voice
		// ID.  Stop the voice ID and start morse ID at
		// 1/2 volume.

		case MSG_AUD_UNSQ:
			Ex_StopTimer(TM_AUDIO_TMR1);
			IO_Set(O_VID_MUTE);		// Mute voice ID
			IO_Clear(O_VIDPLAY);		// Stop voice ID output

			// If we are doing a voice ID, complete it in Morse.

			if(!EEP_IDModeIsCW())
			{
				Ex_SendMsgExt(SM_MORSE,
					MSG_MORSE_SRAM, (word)EEP_GetCWIDString());
				Transition_State_B();	// Wait for morse to complete.
			}

			// If we are doing an announcement, skip it.

			else
				Transition_State_A();
//				DoNextTask();
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Audio State D: Wait here while voice ID or announcement
	// is playing.

static void Transition_State_D(void)
{
	Ex_StartEventTimer(
		TM_AUDIO_TMR1, TP_1MS, EEP_GetVoiceIDDuration(), EV_AUD_TMR1);

	IO_Clear(O_VID_MUTE);		// Unmute voice ID

	// Select next state = C
	SET_STATE(State_D, 'D');
}

static void State_D(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Come here when duration timer times out.
		case EV_AUD_TMR1:
			IO_Set(O_VID_MUTE);		// Mute voice ID
			IO_Clear(O_VIDPLAY);		// Stop voice ID output
			DoNextTask();
			break;

		// Here, someone is trying to talk over the voice
		// ID.  Stop the voice ID and start morse ID at
		// 1/2 volume.

		case MSG_AUD_UNSQ:
			Ex_StopTimer(TM_AUDIO_TMR1);
			IO_Set(O_VID_MUTE);		// Mute voice ID
			IO_Clear(O_VIDPLAY);		// Stop voice ID output

			// If we are doing a voice ID, complete it in Morse.

			if(!EEP_IDModeIsCW())
			{
				Ex_SendMsgExt(SM_MORSE,
					MSG_MORSE_SRAM, (word)EEP_GetCWIDString());
				Transition_State_B();	// Wait for morse to complete.
			}

			// If we are doing an announcement, skip it.

			else
				Transition_State_A();
//				DoNextTask();
			break;

		default:
			break;
	}
}

#endif // VOICE_ID

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Audio State E: Delay before Courtesy Beep

static void Transition_State_E(void)
{
	// Start the Courtesy Beep Timer

	Ex_StartEventTimer(
		TM_AUDIO_TMR1, TP_1MS, EEP_GetCBDelay(), EV_AUD_TMR1);

	// Select next state = E
	SET_STATE(State_E, 'E');
}

static void State_E(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Start the Courtesy Beep.
		case EV_AUD_TMR1:
			Transition_State_F();
			break;

		// Cancel the Courtesy Beep.
		case MSG_AUD_UNSQ:
			Ex_StopTimer(TM_AUDIO_TMR1);
			Transition_State_A();
//			DoNextTask();
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Audio State F: Play a Courtesy Beep

static void Transition_State_F(void)
{
	// Start the Courtesy Beep Timer

	Ex_StartEventTimer(
		TM_AUDIO_TMR1, TP_1MS, T_CB_DURATION, EV_AUD_TMR1);

	FGen_SetFreq(FGW_TONE, (EEP_GetCourtesyBeepFreq()*10));
	FGen_SetGain(FGW_TONE, EEP_GetCourtesyBeepLevel());
	FGen_Start(FGW_TONE);

	// Select next state = E
	SET_STATE(State_F, 'F');
}

static void State_F(MSG *pEvent)
{
	DoCommonStateProcessing(pEvent);

	switch(pEvent->id)
	{
		// Stop the Courtesy Beep.
		case EV_AUD_TMR1:
			FGen_Stop(FGW_TONE);
			DoNextTask();
			break;

		// Cancel the Courtesy Beep.
		case MSG_AUD_UNSQ:
			FGen_Stop(FGW_TONE);
			Ex_StopTimer(TM_AUDIO_TMR1);
			Transition_State_A();
//			DoNextTask();
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// DoNextTask: Check audio activity flags, and if any are
// set, send the appropriate event to the Morse State Machine
// or start the voice output.

static void DoNextTask(void)
{
	// Here send the transmitter timeout message
	if(tasks & AU_DO_TO)
	{
		DoMorseFlashMsg(AU_DO_TO, (char *)TOString);
		Transition_State_B();		// Wait for morse to complete.
	}

	// Here send the DTMF error message
	else if(tasks & AU_DO_ERR)
	{
		DoMorseFlashMsg(AU_DO_ERR, (char *)DTMFErrString);
		Transition_State_B();		// Wait for morse to complete.
	}

	// Here send the DTMF OK message.
	else if(tasks & AU_DO_OK)
	{
		DoMorseFlashMsg(AU_DO_OK, (char *)DTMFOKSting);
		Transition_State_B();		// Wait for morse to complete.
	}

#ifdef VOICE_ID
	// Here send the Voice Mode Announcement.
	else if(tasks & AU_DO_ANNO)
	{
		Audio_Cancel(AU_DO_ANNO);		// Clear the flag

		if((EEP_GetVoiceIDDuration() != 0) && (bUnsq == FALSE))
			Transition_State_C();		// Output announcement.
		else
			DoNextTask();			// Recurse
	}
#endif	// VOICE_ID

	// Here send the Callsign ID message.
	else if(tasks & AU_DO_ID)
	{
		Audio_Cancel(AU_DO_ID);			// Clear the flag

#ifdef VOICE_ID

		if(!EEP_IDModeIsCW() && (bUnsq == FALSE))
		{
			if(EEP_GetVoiceIDDuration() != 0)
				Transition_State_C();		// Output voice ID.

			// If no voice recorded, tell user.
			else
			{
				Ex_SendMsgExt(
					SM_MORSE, MSG_MORSE_FLASH, (word)NVString);
				Transition_State_B();		// Wait for morse to complete.
			}
		}
		else

#endif	// VOICE_ID
		{
			Ex_SendMsgExt(
				SM_MORSE, MSG_MORSE_SRAM, (word)EEP_GetCWIDString());
			Transition_State_B();		// Wait for morse to complete.
		}
	}

	// Here do the courtesy beep (unless it's disabled).

	else if((tasks & AU_DO_CB))
	{
		Audio_Cancel(AU_DO_CB);		// Clear the flag
		Transition_State_E();
	}

	// Here, nothing to do.  Audio output is complete.

	else
		AudioComplete();
}

static void AudioComplete(void)
{
	Ex_SendEvent(SM_TX, EV_AUD_CMPLT);	// Tell Tx we are done.
	Transition_State_A();		// Go Idle.
}

static void DoMorseFlashMsg(byte flag, char *pString)
{
	Audio_Cancel(flag);			// Clear the flag
	Ex_SendMsgExt(SM_MORSE, MSG_MORSE_FLASH, (word)pString);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Process events and messages common to all states

static void DoCommonStateProcessing(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// Enable PL Tone Generation.

		case MSG_AUD_UNSQ:
			Audio_Cancel(AU_DO_CB);		// Reset the flag.
			bUnsq = TRUE;
			break;

		// Disable PL Tone Generation.

		case MSG_AUD_SQ:
			bUnsq = FALSE;
			break;

//		case EV_AUD_REVBURST:
//			FGen_Stop(FGW_CTCSS);		// Turn off CTCSS encoding
//			break;

		default:
			break;
	}
}

