/***************************************************************
*																					*
*	avrapex.c: Event Processing Executive for Small Avrs			*
*																					*
*	FIFO Version																*
*																					*
*	Copyright 2006 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************

							*** HISTORY ***

	060903 ALP: Seperated from PSE-508 and made stand-alone.

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "avrapex.h"
#include "app.h"

#ifdef DEBUG
#include "cdebug.h"
#endif

/***************************************************************/

//		*** MACROS ****

/***************************************************************/

//		*** GENERAL DEFINITIONS ****


/***************************************************************/

//		*** DATA STRUCTURE Definitions ***

typedef struct TAG_MSG_BUF
{
	byte count;		// Number of messages in queue.
	byte xOpen;		// Index to first unused event;
	byte xFirst;	// Index to first used event;
	byte size;		// Size of buffer in messages

	MSG *pBuf;		// Pointer to message buffer
} MSG_BUF;


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void DeliverMsg(MSG_BUF *pMsg);
static void SendMsg(MSG_BUF *pBuf, byte dest, byte type, word wData);
//static bool DeleteMsg(MSG_BUF *pBuf, byte type);


/***************************************************************/

//		*** VARIABLES ***

static MSG eventBuf[NUM_EVENTS];	// Buffer of events
static MSG msgBuf[NUM_MSGS];			// Buffer of messages

static MSG_BUF events =
	{0, 0, 0, NUM_EVENTS, eventBuf};		// Events

static MSG_BUF msgs =
	{0, 0, 0, NUM_MSGS, msgBuf};			// Messages


static MSG *pCurrEvent;	// Pointer to event being processed
static byte currentTask;	// Task number of current task.

static unsigned long systemTime;	// System clock.  Counts mS up.

	// Task state pointers.

static void (*tasks[NUM_TASKS])(MSG *pEvent);

	// Timers

static TIMER timers[NUM_TIMERS];

/***************************************************************/

//		*** PROGRAM ***


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Fatal Error Detected

void Ex_FatalError(char *file, int line, byte data)
{
	cli();	// non-nested disable interrupts
	App_FatalError(file, line, data);
	while(1);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void Ex_SendEvent(byte dest, byte id)
{
	Ex_SendEventExt(dest, id, (word)0);
}

void Ex_SendEventExt(byte dest, byte id, word wData)
{
	SendMsg(&events, dest, id, wData);
}

void Ex_SendMsg(byte dest, byte id)
{
	Ex_SendMsgExt(dest, id, (word)0);
}

void Ex_SendMsgExt(byte dest, byte id, word wData)
{
	SendMsg(&msgs, dest, id, wData);
}

static void SendMsg(MSG_BUF *pMsgs, byte dest, byte id, word wData)
{
	MSG *pEvent;
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	if(pMsgs->count++ <= pMsgs->size)
	{
		pEvent = &pMsgs->pBuf[pMsgs->xOpen];
		pEvent->dest = dest;
		pEvent->id = id;
		pEvent->wData = wData;

		if(++pMsgs->xOpen >= pMsgs->size)
			pMsgs->xOpen = 0;
	}
	else
		FATAL;

	EI(ieStatus);		// Nested enable interrupts
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Read the system clock:

unsigned long Ex_GetSystemTime(void)
{
	return(systemTime);
}

	// Start a polling timer.  This timer will call the
	// service address upon expriation, and will automatically
	// reload.  It counts in units of mS and may be set for up
	// to 65535 mS.

void Ex_StartPollTimer(
	byte timerNum, word duration, void (*pServiceFunc)(void))
{
	byte ieStatus;
	TIMER *pTimer;

	DI(ieStatus);		// Nested disable interrupts

	pTimer = &timers[timerNum];

	pTimer->type = TT_POLL_TIMER;
	pTimer->state = TS_RUNNING;
	pTimer->prescaler = pTimer->resetPrescaler = 1;
	pTimer->timeCount = pTimer->resetTime = duration;
	pTimer->pServiceFunc = pServiceFunc;

	EI(ieStatus);		// Nested enable interrupts
}

	// Start an event timer.  This timer will send an event
	// upon expriation and will not reload.  It counts in units
	// indicated by the prescaler.

void Ex_StartEventTimer(
	byte timerNum, word prescaler, word duration, byte id)
{
	byte ieStatus;
	TIMER *pTimer;

	DI(ieStatus);		// Nested disable interrupts

	pTimer = &timers[timerNum];

	pTimer->type = TT_EVENT_TIMER;
	pTimer->state = TS_RUNNING;
	pTimer->prescaler = pTimer->resetPrescaler = prescaler;
	pTimer->timeCount = pTimer->resetTime = duration;

	pTimer->dest = currentTask;
	pTimer->id = id;

	EI(ieStatus);		// Nested enable interrupts
}


	// Stop the timer indicated by the timerNum parameter.
	// If the timer is an event timer, check to see if it
	// is timed-out and if so, search the event queue for
	// the event and delete it also.

void Ex_StopTimer(byte timerNum)
{
	byte i;
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

	if(timers[timerNum].type == TT_EVENT_TIMER)
	{
		if(timers[timerNum].state == TS_TIMED_OUT)
		{
			// Delete the timer's events.

			if(events.count != 0)
			{
				i = events.xFirst;		// Point at first event.

				while(i != events.xOpen)
				{
					if(events.pBuf[i].id == timers[timerNum].id)
						events.pBuf[i].id = EV_NULL; // Event found.

					if(++i >= events.size)		// Queue wrap-around
						i = 0;
				}
			}
		}
	}

	timers[timerNum].state = TS_STOPPED;

	EI(ieStatus);		// Nested enable interrupts
}

bool Ex_TimerIsRunning(byte timerNum)
{
	return(timers[timerNum].state == TS_RUNNING);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set state machine execution address

void Ex_SetState(void (*pState)(MSG *pEvent))
{
	tasks[currentTask] = pState;
}


void Ex_InitTask(byte taskNum, void (*pInitAddr)(void))
{
	currentTask = taskNum;
	(*pInitAddr)();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Main:

//	This function is passed control by the C start-up
//	file.  It performs power-up initialization, then
//	goes into a loop in which it retreives an event
//	and distributes the event to the proper state
//	machine.

void main(void)
{
	App_InitIO();			// Initialize application I/Os

	// Initialize the operating system:

	events.pBuf = eventBuf;
	msgs.pBuf = msgBuf;
	// systemTime = 0;				// Actually done by C init.

	// Timer 0 Initialization: This timer is used for general
	// purpose software timer generation.  It is initialized for
	// a 1 ms timebase.

	TCNT0 = TMR0_COUNT;	// Set initial count

	// Select 1/64 prescaler and start timer.

#ifdef __AVR_ATmega168__
	TCCR0B = TMR0_START;
#elif defined __AVR_ATmega8__
	TCCR0 = TMR0_START;
#endif

	// Init Timer Interrupts

#ifdef __AVR_ATmega168__
	TIMSK0 |= BV(TOIE0);
#elif defined __AVR_ATmega8__
	TIMSK |= BV(TOIE0);
#endif

	App_InitTasks();			// Initialize the application tasks

	sei();	// non-nested enable interrupts

	FOREVER
	{
		cli();	// non-nested disable interrupts

		// Check to see if there are any messages to deliver.
		// Messages from the "msgs" queue are high priority,
		// messages from the "events" queue are low priority.

		if(msgs.count != 0)
			DeliverMsg(&msgs);
		else
		{
			if(events.count != 0)
				DeliverMsg(&events);
		}

		sei();	// non-nested enable interrupts
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Deliver a message from one of the message queues.

static void DeliverMsg(MSG_BUF *pMsgs)
{

	pCurrEvent = &pMsgs->pBuf[pMsgs->xFirst];
	currentTask = pCurrEvent->dest;

	if(pCurrEvent->id != EV_NULL)
	{
		sei();	// non-nested enable interrupts
		(*tasks[currentTask])(pCurrEvent);
		cli();	// non-nested disable interrupts
	}

	if(++pMsgs->xFirst >= pMsgs->size)
		pMsgs->xFirst = 0;

	pMsgs->count--;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//	Timer 0 Overflow Handler

//	This interrupt generates general purpose time bases for
//	software generated timers.  Time bases of 1, 10, and
//	100 ms, as well as 1 and 10 seconds are generated.
//	Software timers may base their counting rates on any
//	of the available time bases.

SIGNAL(SIG_OVERFLOW0)
{
//DB_Strobe();

	byte i;
	TIMER *pTimer;

	TCNT0 += TMR0_COUNT;	// Refresh the timer

	systemTime++;			// Increment system time.

	// Enable interrupts during this interrupt.  This
	// allows higher priority interrupts such as the
	// audio generator to execute.

	sei();	// non-nested enable interrupts

	for(i=0, pTimer = &timers[0]; i<NUM_TIMERS; i++, pTimer++)
	{
		if(pTimer->state == TS_RUNNING)
		{
			if(--pTimer->prescaler == 0)
			{
				pTimer->prescaler = pTimer->resetPrescaler;
				if(--pTimer->timeCount == 0)
				{
					if(pTimer->type == TT_POLL_TIMER)
					{
						pTimer->timeCount = pTimer->resetTime;
						(*pTimer->pServiceFunc)();
					}

					else if(pTimer->type == TT_EVENT_TIMER)
					{
						pTimer->state = TS_TIMED_OUT;
						Ex_SendEvent(pTimer->dest, pTimer->id);
					}

					else
						FATAL;
				}
			}
		}
	}

//DB_Strobe();

	cli();	// non-nested disable interrupts
}

