/***************************************************************
*																					*
*	bitio.c: Bit I/O Routines												*
*																					*
*	This file contains bit level hardware interface functions	*
*	to manipulate the inputs and outputs contained on the bit	*
*	addressable I/O devices U11, U12, and .							*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/


//		*** INCLUDE FILES ***

#include <avr/interrupt.h>

#include "avrstd.h"
#include "app.h"
#include "bitio.h"
#include "avrapex.h"
#include "cdebug.h"

//		*** MACROS ****


//		*** GENERAL DEFINITIONS ****


//	Data Direction Registers

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))
#	define PORTB_DIR	0b10001110	// Port B DDR
#	define PORTB_INI	0b00111110	// Initial State
#else // PCB_VERSION == 21
#	define PORTB_DIR	0b10011010	// Port B DDR
#	define PORTB_INI	0b00101010	// Initial State
#endif
					// PB.0: PL_DET - Input
					// PB.1: RX_MUTE - Output
					// PB.2: PCB 1.0: CWIDENA - Output
					//			PCB 2.0: Unused - Output
					//			PCB 2.1: DTMF_DET - Input
					// PB.3: CWID - Output
					// PB.4: Debug Output (S1B - Input)
					// PB.5: S1A - Input
					// PB.6: XTAL1 - Input
					// PB.7: XTAL2 - Output

					// Initial State:
					// RX_MUTE high (muted).
					// CWIDENA high (disabled).
					// CWID high.
					// PB5 pullups for S1.

#if PCB_VERSION == 10
#	define PORTC_DIR	0b00101111	// Port C Data Dir.
#	define PORTC_INI	0b00000001	// Initial State
#else // PCB_VERSION == 20 or 21
#	define PORTC_DIR	0b00101011	// Port C Data Dir.
#	define PORTC_INI	0b00000001	// Initial State

#endif
					// PC.0: VIDREC - Output
					// PC.1: UNSQUELCHLED - Output
					// PC.2: PCB 1.0: DTMFLED - Output
					//			PCB 2.0: CTCSS_DET_A - Input
					//			PCB 2.1: CTCSS_DET_A - Input
					// PC.3: VID_MUTE - Output
					// PC.4: SQUELCH - Input
					// PC.5: PTT - Output
					// PC.6: RESET - Input

					// Initial State:
					// VIDREC high (off),
					// LEDs low (off),
					// VID_MUTE low (muted)
					// PTT low (off).

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))
#	define PORTD_DIR	0b00000110	// Port D Data Dir.
#	define PORTD_INI	0b00000111	// Initial State

#else // PCB_VERSION == 21
#	define PORTD_DIR	0b00001110	// Port D Data Dir.
#	define PORTD_INI	0b00000111	// Initial State
#endif
					// PD.0: RxD - Input
					// PD.1: TxD - Output
					// PD.2: VIDPLAY - Output
					// PD.3: PCB 1.0: DTMF_DET - Input
					//			PCB 2.0: DTMF_DET - Input
					//			PCB 2.1: CTCSS_ENC_PWM - Output
					// PD.4: DTMF0 - Input
					// PD.5: DTMF1 - Input
					// PD.6: DTMF2 - Input
					// PD.7: DTMF3 - Input

					// Initial State
					// TxD=1, RxD PUP =1
					// VIDPLAY high (off)
					// PD3-PD7 inputs.

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

/***************************************************************/

//		*** VARIABLES ***

#ifdef IO_DEBUG

//	The following variables are used to simulate bit-addressable
//	I/O on the AVR simulator.

static byte db_IN = 0xFF;	// Bit-Addressable Inputs
static byte db_OUT;			// Bit-Addressable Outputs

#endif

/***************************************************************/

//	PortInit:

//	This function initializes the microprocessor I/O ports
//	on reset.

void IO_Init(void)
{
	DDRB = PORTB_DIR;
	PORTB = PORTB_INI;	// Init Port B State:

	DDRC = PORTC_DIR;
	PORTC = PORTC_INI;	// Init Port C State:

	DDRD = PORTD_DIR;
	PORTD = PORTD_INI;	// Init Port D State:
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	SetIO:

// This function sets a logical output to the active state.
// It performs both the logical to physical output translation
// as well as the active level translation.  All logical outputs
// are assumed to be active high.

void IO_Set(byte bitadr)
{
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

#ifdef IO_DEBUG

	db_OUT |= (1<<bitadr);

#else
	switch(bitadr)
	{
		case O_VIDPLAY:
			cbi(PORTD,2);	// VIDPLAY: Port D bit 2, active low.
			break;

#if PCB_VERSION == 10
		case O_CWIDENA:
			cbi(PORTB,2);	// CWIDENA: Port B bit 2, active low.
			break;
#endif

		case O_RXENA:
			cbi(PORTB,1);	// RX_MUTE: Port B bit 1, active high.
			break;			// (Clear to enable Rx Audio)

		case O_PTT:
			sbi(PORTC,5);	// PTT: Port C bit 5, active high
			break;

		case O_VID_MUTE:
			sbi(PORTC,3);	// VID_MUTE: Port C bit 3, active high
			break;

		case O_UNSQULED:
			sbi(PORTC,1);	// UNSQUELCHED: Port C bit 1, active high
			break;

#if PCB_VERSION == 10
		case O_DTMFLED:
			sbi(PORTC,2);	// DTMFLED: Port C bit 2, active high
			break;
#endif

		case O_VIDREC:
			cbi(PORTC,0);	// VIDREC: Port C bit 0, active low
			break;

		default:
			FATAL;
			break;
	}

#endif
	EI(ieStatus);		// Nested enable interrupts
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	ClearIO:

// This function clears an output port to the inactive state.
// It performs both the logical to physical output translation
// as well as the active level translation.  All logical outputs
// are assumed to be active high.

void IO_Clear(byte bitadr)
{
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

#ifdef IO_DEBUG

	db_OUT &= ~(1<<bitadr);

#else
	switch(bitadr)
	{
		case O_VIDPLAY:
			sbi(PORTD,2);	// VIDPLAY: Port D bit 2, active low.
			break;

#if PCB_VERSION == 10
		case O_CWIDENA:
			sbi(PORTB,2);	// CWIDENA: Port B bit 2, active low.
			break;
#endif

		case O_RXENA:
			sbi(PORTB,1);	// RX_MUTE: Port B bit 1, active high
			break;			// (Set to disable Rx Audio)

		case O_PTT:
			cbi(PORTC,5);	// PTT: Port C bit 5, active high
			break;

		case O_VID_MUTE:
			cbi(PORTC,3);	// VID_MUTE: Port C bit 3, active high
			break;

		case O_UNSQULED:
			cbi(PORTC,1);	// UNSQUELCHED: Port C bit 1, active high
			break;

#if PCB_VERSION == 10
		case O_DTMFLED:
			cbi(PORTC,2);	// DTMFLED: Port C bit 2, active high
			break;
#endif

		case O_VIDREC:
			sbi(PORTC,0);	// VIDREC: Port C bit 0, active low
			break;

		default:
			FATAL;
			break;
	}

#endif
	EI(ieStatus);		// Nested enable interrupts
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// GetIO:

//	This function reads a bit-addressable input port.  All inputs
// are converted to active TRUE before this function returns.

bool IO_Get(byte bitadr)
{
	byte ieStatus;
	volatile byte bit;

	DI(ieStatus);		// Nested disable interrupts

#ifdef IO_DEBUG
	if(db_IN & bitadr)
		bit = TRUE;
	else
		bit = FALSE;

#else

	switch(bitadr)
	{
		case I_DTMF_DET:
#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))
			bit = PIND & 0b00001000;	// Active high
#else // PCB_VERSION == 21
			bit = PINB & 0b00000100;
#endif
			break;

		case I_S1A:
			bit = ~(PINB | 0b11011111); // Active low
			break;

		case I_S1B:
			bit = ~(PIND | 0b11101111); // Active low
			break;

		case I_SQUELCH:
			bit = PINC & 0b00010000;
			break;

		default:
			FATAL;
			break;
	}

#endif

	EI(ieStatus);		// Nested enable interrupts
	return(bit != 0);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	This routine reads the DTMF bits and returns them to
//	the caller in bits 0 through 3 of the returned byte.
// ALP 060319 Modified for PSE-508

byte IO_GetDTMFCode(void)
{
	byte code;
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

#ifndef IO_DEBUG
	code = (PIND & 0xF0) >> 4;

#else
	code = (db_IN & 0xF0) >> 4;

#endif

	EI(ieStatus);		// Nested enable interrupts
	return(code);
}

