/***************************************************************
*																					*
*	dtmf.c: DTMF Command Recognition & Execution State Machine	*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************

						DTMF Keyboard Codes

							* 0x0B
							# 0x0C
							0 0x0A
							1 0x01
							2 0x02
							3 0x03
							4 0x04
							5 0x05
							6 0x06
							7 0x07
							8 0x08
							9 0x09
							A 0x0D
							B 0x0E
							C 0x0F
							D 0x00

	DTMF Commands:

	Command										Keys(s)
	Start of DTMF command					C

	Turn Transmitter Off:					C00
	Turn Transmitter On:						C01

	Disable PL Detection:					C10
	Enable PL Detection:						C11

	Disable PL Encode Announcements:		C12
	Enable PL Encode Announcements:		C13

	Disable Courtesy Beep					C14
	Enable Courtesy Beep						C15

	Mute DTMF Codes							C16
	Unmute DTMF Codes							C17

	Disable Reverse Burst					C18
	Enable Reverse Burst						C19

	Enable Voice ID Recording:				C2

	Set ID / Announcement Mode:			C3X	(X = Mode)

	Set CW ID String							C4 <string>

	Force an ID									C5

	Log Off										C6

	Set PL Encode Mode						C7X	(X = Mode)

	Sync Announcement Timer:		B		(Following password entry)
	Play Announcement					A		(W/O password entry)

****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "dtmf.h"
#include "bitio.h"
#include "morse.h"
#include "serial.h"
#include "audio.h"
#include "parse.h"
#include "plsm.h"

#ifdef DEBUG
#include "cdebug.h"
#endif

	// undef define

#ifdef DTMF_DEBUG
#undef	DTMF_DEBUG_UC_STATE_FLAGS	// Show state flags in UC
#define	DTMF_DEBUG_LC_STATE_FLAGS	// Show state flags in LC
#define	DTMF_OTHER_STATE_FLAGS		// Show other flags in LC
#undef	DTMF_OTHER_STATE_DATA		// Show other flags in LC
#undef	DTMF_DEBUG_EV_DISP			// Show event messages
#endif

/***************************************************************/

//		*** MACROS ****

#if defined DTMF_DEBUG_UC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y))
#elif defined DTMF_DEBUG_LC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y)|0x20)
#else
#define SET_STATE(X,Y)	Ex_SetState(X)
#endif

/***************************************************************/

//		*** GENERAL DEFINITIONS ****


//		DTMF Keyboard Codes

#define DTMF_STAR		0x0B		//	* 0x0B
#define DTMF_SHARP	0x0C		//	# 0x0C
#define DTMF_0			0x0A		//	0 0x0A
#define DTMF_1			0x01		//	1 0x01
#define DTMF_2			0x02		//	2 0x02
#define DTMF_3			0x03		//	3 0x03
#define DTMF_4			0x04		//	4 0x04
#define DTMF_5			0x05		//	5 0x05
#define DTMF_6			0x06		//	6 0x06
#define DTMF_7			0x07		//	7 0x07
#define DTMF_8			0x08		//	8 0x08
#define DTMF_9			0x09		//	9 0x09
#define DTMF_A			0x0D		//	A 0x0D
#define DTMF_B			0x0E		//	B 0x0E
#define DTMF_C			0x0F		//	C 0x0F
#define DTMF_D			0x00		//	D 0x00

	//	DTMF Command Codes

#define DTMF_CMD_STRT	DTMF_C	// Start of DTMF command
#define DTMF_ALT_STRT	DTMF_SHARP // Alternate start of DTMF command

#define DTMF_OFF			DTMF_0	// Turn something off.
#define DTMF_ON			DTMF_1	// Turn something on.

#define DTMF_TXENA		DTMF_0	// Enable/Disable Transmitter
#define DTMF_FLAGS		DTMF_1	// Enable/Disable Certian Features
#define DTMF_REC			DTMF_2	// Enable Voice ID recording.
#define DTMF_VID			DTMF_3	// Select VID Mode
#define DTMF_ENTER_CW	DTMF_4	// Enter CW ID String

#define DTMF_FORCE_ID	DTMF_5	// Force transmitter to send ID.
#define DTMF_LOG_OFF		DTMF_6	// Log Off Immediately

#define DTMF_ECTCSS		DTMF_7	// Enter CTCSS Encode Mode

#ifdef PL_TWEEK
#	define DTMF_PLTWEEK		DTMF_9	// Tweek PL Parameters
#endif

	// Timeouts

#define T_DTMF_PASS_TO	60			// 60 second password timeout.

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

//static void DTMFComplete(void);
//static void EndWithError(void);
static void DTMF_CheckISR(void);
static byte MapDTMFCode(byte code);

static void Transition_State_A(void);
static void State_A(MSG *pEvent);
static void Transition_State_B(void);
static void State_B(MSG *pEvent);
static void Transition_State_C(void);
static void State_C(MSG *pEvent);
static void Transition_State_D(void);
static void State_D(MSG *pEvent);
static void Transition_State_E(void);
static void State_E(MSG *pEvent);
static void Transition_State_F(void);
static void State_F(MSG *pEvent);
//static void Transition_State_G(void);
//static void State_G(MSG *pEvent);
static void Transition_State_H(void);
static void State_H(MSG *pEvent);

static void ProcessTxPwrCmd(void);
static void ProcessFlagsCmd(void);

#ifdef VOICE_ID
static void ProcessVIDCmd(void);
#endif

#ifdef CTCSS_ENCODING
static void ProcessECTCSSCmd(void);
#endif

/***************************************************************/

//		*** VARIABLES ***


static bool prevDTMF;		// ISR DTMF state on previous pass

static byte digitCount;		// Counter for digits received.
static byte command;			// Command character (received after "C").
static byte dataByte1;		// First data byte after command.
static byte cwCharIndex;	// Number of whole cw characters entered.
static char cwString[Z_WORDSTRING];	// CWID string during entry.

static bool oddCwChar;		// Entering odd half of cw character
static bool passEntered;	// Password was entered if true.

//		*** CONSTANTS ***

	// Use this table to map the received DTMF code to the
	// numeric code used in the firmware.

static prog_uchar DTMFMappingTbl[] =
{
		//	Firmware		DTMF
		//	code			code

			'D',		// 0x00 = DTMF_D
			'1',		// 0x01 = DTMF_1
			'2',		// 0x02 = DTMF_2
			'3',		// 0x03 = DTMF_3
			'4',		// 0x04 = DTMF_4
			'5',		// 0x05 = DTMF_5
			'6',		// 0x06 = DTMF_6
			'7',		// 0x07 = DTMF_7
			'8',		// 0x08 = DTMF_8
			'9',		// 0x09 = DTMF_9
			'0',		// 0x0A = DTMF_0
			'*',		// 0x0B = DTMF_STAR
			'#',		// 0x0C = DTMF_SHARP
			'A',		// 0x0D = DTMF_A
			'B',		// 0x0E = DTMF_B
			'C'		// 0x0F = DTMF_C
};

/***************************************************************/

//		*** API ***

// Perform reset initialization of DTMF detector

void DTMF_Init(void)
{
	Ex_StartPollTimer(TM_DTMF_CHECK, 10, DTMF_CheckISR);

	prevDTMF = FALSE;
	Transition_State_F();
}

static byte MapDTMFCode(byte code)
{
	return(PRG_RDB((byte *)DTMFMappingTbl + code));
}

	// Restart the password timeout.

static void RestartPassTimer(void)
{
	Ex_StopTimer(TM_DTMF_PASS_TIMEOUT);
	Ex_StartEventTimer(TM_DTMF_PASS_TIMEOUT, TP_SEC,
		T_DTMF_PASS_TO, EV_DTMF_PASS_TO);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Check for new DTMF code received.  Called by
// Timer 0 interrupt service routine.

// ALP 060525: Checked to see if interrupts should be disabled
//					around this routine.  It appears that they
//					do not need to be.

static void DTMF_CheckISR(void)
{
	if(IO_Get(I_DTMF_DET))
	{
		// Here, DTMF Detected

		if(!prevDTMF)			// Was previous state low?
		{
			prevDTMF = TRUE;	// make prevous state high.
			Ex_SendEventExt(SM_DTMF, EV_DTMF_NEW, IO_GetDTMFCode());
		}
	}

		// Here, DTMF Detected Flag is Low

	else
		prevDTMF = FALSE;		// make prevous state low.
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Enable or disable DTMF detection

void DTMF_Enable(void)
{
	Ex_SendMsg(SM_DTMF, MSG_DTMF_ENA); // Enable DTMF detection
}

void DTMF_Disable(void)
{
	Ex_SendMsg(SM_DTMF, MSG_DTMF_DIS); // Disable DTMF detection
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	DTMF state F: Wait here until DTMF detection is enabled.

static void Transition_State_F(void)
{
	// Select next state = F
	SET_STATE(State_F, 'F');
#if PCB_VERSION == 10
	IO_Clear(O_DTMFLED);	// Turn off the LED.
#endif
}

static void State_F(MSG *pEvent)
{
	switch(pEvent->id)
	{
		case MSG_DTMF_ENA:
			Transition_State_A();	// Enable DTMF detection.
			break;

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;
#ifdef DTMF_OTHER_STATE_FLAGS
			Serial_TxChar('-');
#endif
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//	DTMF state A: Receive the password.

static void Transition_State_A(void)
{
	SET_STATE(State_A, 'A'); // Select next state = A
}


static void State_A(MSG *pEvent)
{
	switch(pEvent->id)
	{
		case MSG_DTMF_DIS:
			Transition_State_F();	// Disable DTMF detection.
			break;

		// Here, a DTMF code was detected.  Check to see if the
		// password was successfully entered.

		case EV_DTMF_NEW:
			// Here, command to play announcement.
			if(((pEvent->wData == DTMF_A)
				|| (pEvent->wData == DTMF_STAR))
				&& EEP_AnnouncementEnabled())
			{
				Audio_Schedule(AU_DO_ANNO);
				Ex_SendMsg(SM_TX, EV_TX_ANNTO);	// Notify Tx Task
				Transition_State_E();				// Wait for unkey.
			}

			else if(passEntered)
			{
				Ex_SendMsg(
					SM_TX, MSG_DTMF_INPROG);	// Say DTMF in progress.
#if PCB_VERSION == 10
				IO_Set(O_DTMFLED);				// Turn on the LED.
#endif
				Audio_Schedule(AU_DO_ERR);		// Say error for now.

				if((pEvent->wData == DTMF_CMD_STRT)
					|| (pEvent->wData == DTMF_ALT_STRT))
					Transition_State_B();	// Wait for DTMF command code.

				// Sync announcement timer
				else if((pEvent->wData == DTMF_B)
					&& EEP_AnnouncementEnabled())
				{
					Audio_StartAnnouncementTimer();
					Audio_Cancel(AU_DO_ERR);		// No error.
					Audio_Schedule(AU_DO_OK);		// Say OK.
					Transition_State_E();			// Else wait for unkey.
				}

				else
					Transition_State_E();	// Else wait for unkey.
			}

			// Here, password has not been entered yet.

			else
			{
#ifdef DTMF_OTHER_STATE_DATA
				Serial_TxHexChar(pEvent->wData);
				Serial_TxChar(' ');
#endif
//				if(MapDTMFCode(pEvent->wData) == 'C')
				if((pEvent->wData == DTMF_CMD_STRT)
					|| (pEvent->wData == DTMF_ALT_STRT))
				{
					Ex_SendMsg(
						SM_TX, MSG_DTMF_INPROG);	// Say DTMF in progress.
#if PCB_VERSION == 10
					IO_Set(O_DTMFLED);				// Turn on the LED.
#endif
					Audio_Schedule(AU_DO_ERR);		// Say error for now.
						Transition_State_C();
				}
			}

			break;

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;

#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif

			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	DTMF State B: Start a timer.  If the timer times out before
	// a command is received, exit DTMF mode.
	// Wait here for the first character in a command.

static void Transition_State_B(void)
{
	SET_STATE(State_B, 'B');	// Select next state = B
}


static void State_B(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// If we get the disable event at this stage we
		// are between commands.

		case MSG_DTMF_DIS:
			Transition_State_F();	// Go Idle.
			break;

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif
			Transition_State_E();	// Wait for unkey.
			break;

		// Here, a DTMF code was detected.

		case EV_DTMF_NEW:
			RestartPassTimer();	// Restart the password timeout.

			command = (byte)pEvent->wData;
			switch(command)
			{
				// Process two character commands here.

				case DTMF_VID:
				case DTMF_TXENA:
				case DTMF_FLAGS:
				case DTMF_ECTCSS:

#ifdef PL_TWEEK
				case DTMF_PLTWEEK:
#endif
					Transition_State_H();	// Get second digit.
					break;

				// Here, a command to open recording of voice ID.

#ifdef VOICE_ID

				case DTMF_REC:
					Ex_SendMsg(SM_RX, MSG_VID_ENA);
					Audio_Cancel(AU_DO_ERR);	// No error.
					Audio_Schedule(AU_DO_OK);
					Transition_State_E();	// Wait for unkey
					break;

#endif // VOICE_ID

				// Allow the user to enter a CW string.

				case DTMF_ENTER_CW:
					Transition_State_D();	// Get characters.
					break;

				// Force the repeater to do an immediate ID.

				case DTMF_FORCE_ID:
					Ex_SendMsg(SM_TX, MSG_TX_IDCMD);
					Audio_Cancel(AU_DO_ERR);	// No error.
					Audio_Schedule(AU_DO_OK);
					Transition_State_E();	// Wait for unkey
					break;

				// Immediately Log off the DTMF mode.

				case DTMF_LOG_OFF:
					passEntered = FALSE;
					Ex_StopTimer(TM_DTMF_PASS_TIMEOUT);
#ifdef DTMF_OTHER_STATE_FLAGS
					Serial_TxChar('-');
#endif
					Audio_Cancel(AU_DO_ERR);	// No error.
					Audio_Schedule(AU_DO_OK);
					Transition_State_E();	// Wait for unkey
					break;

				// Here, command number unknown.
				default:
					Transition_State_E();	// Wait for unkey
					break;
			}
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	DTMF State C: Decode first, second, and third characters
	// of password.

static void Transition_State_C(void)
{
	passEntered = TRUE;	// Default to properly entered.
	digitCount = 0;		// Count the digits.

	// Select next state = C
	SET_STATE(State_C, 'C');
}

static void State_C(MSG *pEvent)
{
	switch(pEvent->id)
	{
		case MSG_DTMF_DIS:
			if(digitCount == (Z_DTMF_PASSWORD-1) && passEntered)
			{
				Audio_Cancel(AU_DO_ERR);	// No error.
				Audio_Schedule(AU_DO_OK);
				RestartPassTimer();	// Restart the password timeout.
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('+');
#endif

			}

			else
				passEntered = FALSE;

			Transition_State_F();		// Go Idle.
			break;

		// Here, collecting password digits.

		case EV_DTMF_NEW:
#ifdef DTMF_OTHER_STATE_DATA
				Serial_TxHexChar(pEvent->wData);
#endif

			if(MapDTMFCode(pEvent->wData) !=
				*(EEP_GetDTMFPassword() + digitCount++))
			{
				passEntered = FALSE;
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif
			}

#ifdef DTMF_OTHER_STATE_DATA
			Serial_TxChar(' ');
#endif
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	DTMF State D: Enter CW ID String

static void Transition_State_D(void)
{
	cwCharIndex = 0;
	oddCwChar = TRUE;

	// Select next state = D
	SET_STATE(State_D, 'D');
}


static void State_D(MSG *pEvent)
{
	char c;

	switch(pEvent->id)
	{
		// If we get the disable event, test for CWID properly
		// entered.

		case MSG_DTMF_DIS:
			if(cwCharIndex != 0 && oddCwChar)
			{
				cwString[cwCharIndex] = '\0';		// Terminate string.
				EEP_SetCWIDString(cwString);		// Put it in EEPROM
				Transition_State_F();				// Go idle.
				Audio_Cancel(AU_DO_ERR);			// No error.
				Audio_Schedule(AU_DO_OK);
			}

			else
				// Here, string index invalid or partial char
				// was entered.
				Transition_State_F();				// Go idle.

			break;

		// Here, a DTMF code was detected.

		case EV_DTMF_NEW:
			RestartPassTimer();	// Restart the password timeout.
			c = MapDTMFCode(pEvent->wData);

			if(c >= '0' && c <= '9')
			{
				if(oddCwChar)
				{
					// Here, odd character.
					oddCwChar = FALSE;
					dataByte1 = c & 0x0F;
					RestartPassTimer();	// Restart the password timeout.
				}

				else
				{
					// Here, even character.

					c = ((dataByte1 * 10) + (c & 0x0F));

					if((c = Morse_GetCharFromTable(c)) == 0x00)
						Transition_State_E();	// Error.

					else
					{
						cwString[cwCharIndex++] = c;

						if(cwCharIndex == Z_WORDSTRING)
							// Here, string too long.
							Transition_State_E();	// Else wait for unkey.

						else
						{
							// Here, string length OK.
							RestartPassTimer();	// Restart the password timeout.
							oddCwChar = TRUE;
						}
					}
				}
			}
			else
				Transition_State_E();	// Wait for unkey

			break;

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;	// Time out the password.
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif
			Transition_State_E();	// Wait for unkey.
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	DTMF State E: The command is complete.  Wait for sender to
// unkey, then send a response message.

static void Transition_State_E(void)
{
	// Select next state = E
	SET_STATE(State_E, 'E');
}


static void State_E(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// If we get the disable event at this stage we
		// are between commands.

		case MSG_DTMF_DIS:
			Transition_State_F();					// Go idle.
			break;

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;	// Time out the password.
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	DTMF State H: Receive a second command number for an on/off
// type command.

static void Transition_State_H(void)
{
	// Select next state = H
	SET_STATE(State_H, 'H');
}


static void State_H(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// If we get the disable event at this stage we have
		// not completed a command and we report an error.

		case MSG_DTMF_DIS:
			Transition_State_F();			// Go idle.
			break;

		// Here, a DTMF code was detected.

		case EV_DTMF_NEW:
			RestartPassTimer();	// Restart the password timeout.

			dataByte1 = (byte)pEvent->wData;

			Audio_Cancel(AU_DO_ERR);	// No error.
			Audio_Schedule(AU_DO_OK);

			switch(command)
			{
				// Process two character commands here.
				case DTMF_TXENA:
					ProcessTxPwrCmd(); // Enable Transmitter
					break;

				case DTMF_FLAGS:
					ProcessFlagsCmd();	// Enable Tone Squelch
					break;

#ifdef VOICE_ID
				case DTMF_VID:
					ProcessVIDCmd();		// Set ID Mode
					break;
#endif // VOICE_ID

#ifdef CTCSS_ENCODING
				case DTMF_ECTCSS:
					ProcessECTCSSCmd();	// Set CTCSS Encode Mode
					break;
#endif // CTCSS_ENCODING

#ifdef PL_TWEEK
				case DTMF_PLTWEEK:
					EEP_SetPLTweek(dataByte1);
					PL_SetToneDecodeThresholds();
					break;
#endif

				// Here, command number has already been vetted,
				// so there is a fatal error.

				default:
					FATAL;
					break;
			}

			Transition_State_E();	// Wait for unkey.
			break;

		// Here, the password timed out.

		case EV_DTMF_PASS_TO:
			passEntered = FALSE;	// Time out the password.
#ifdef DTMF_OTHER_STATE_FLAGS
				Serial_TxChar('-');
#endif
			Transition_State_E();	// Wait for unkey.
			break;

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Process a command to turn the transmitter on or off.

static void ProcessTxPwrCmd(void)
{
	if(dataByte1 == DTMF_OFF)
	{
		// Turn transmitter off

		Ex_SendMsg(SM_TX, MSG_TX_OFF);
#ifdef DTMF_OTHER_STATE_FLAGS
		Serial_TxChar('t');	// %% debug
#endif
	}

		// Else turn it on.
	else if(dataByte1 == DTMF_ON)
	{
		EEP_SetTxEnableState(TRUE);

#ifdef DTMF_OTHER_STATE_FLAGS
		Serial_TxChar('T');	// %% debug
#endif
	}

	else
	{
		Audio_Cancel(AU_DO_OK);		// Flag an error.
		Audio_Schedule(AU_DO_ERR);
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Process a command to turn certian features on or off.

static void ProcessFlagsCmd(void)
{
	switch(dataByte1)
	{
		// Turn CTCSS detection off
		case DTMF_0:
			EEP_SetPLDecodeState(FALSE);
#ifdef DTMF_OTHER_STATE_FLAGS
			Serial_TxChar('p');	// %% debug
#endif
			break;

		// Turn CTCSS Detection on
		case DTMF_1:
			EEP_SetPLDecodeState(TRUE);
#ifdef DTMF_OTHER_STATE_FLAGS
			Serial_TxChar('P');	// %% debug
#endif
			break;

		// Turn CTCSS Encode Announcements off

#ifdef CTCSS_ENCODING
//		case DTMF_2:
//			EEP_SetPLEncodeAnnouncements(FALSE);
//			break;

		// Turn CTCSS Encode Announcements on

//		case DTMF_3:
//			EEP_SetPLEncodeAnnouncements(TRUE);
//			break;
#endif

		// Turn Courtesy Beep off

		case DTMF_4:
			EEP_SetCourtesyBeepState(FALSE);
			break;

		// Turn Courtesy Beep on

		case DTMF_5:
			EEP_SetCourtesyBeepState(TRUE);
			break;

		// Mute DTMF Codes

		case DTMF_6:
			EEP_SetDTMFMuteState(TRUE);
			break;

		// Unmute DTMF Codes

		case DTMF_7:
			EEP_SetDTMFMuteState(FALSE);
			break;

#ifdef CTCSS_ENCODING
		// Disable Reverse Burst
		case DTMF_8:
			EEP_SetReverseBurstState(FALSE);
			break;

		// Enable Reverse Burst
		case DTMF_9:
			EEP_SetReverseBurstState(TRUE);
			break;
#endif

		default:
			Audio_Cancel(AU_DO_OK);		// Flag an error.
			Audio_Schedule(AU_DO_ERR);
			break;
	}
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Process a command to set ID mode byte:

#ifdef VOICE_ID

static void ProcessVIDCmd(void)
{
	switch(dataByte1)
	{
		case DTMF_0:
			EEP_SetIDMode(MODE_ID_CW);
			break;

		case DTMF_1:
			EEP_SetIDMode(MODE_ID_VOICE);
			break;

		case DTMF_2:
			EEP_SetIDMode(MODE_ID_TIMED_ANN);
			break;

		case DTMF_3:
			EEP_SetIDMode(MODE_ID_MANUAL_ANN);
			break;

		default:
			Audio_Cancel(AU_DO_OK);		// Flag an error.
			Audio_Schedule(AU_DO_ERR);
			break;
	}
}

#endif // VOICE_ID

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Process a command to set the CTCSS Encoder mode:

#ifdef CTCSS_ENCODING

static void ProcessECTCSSCmd(void)
{
	switch(dataByte1)
	{
		case DTMF_0:
			EEP_SetCTCSSEncodeMode(MODE_ECTCSS_OFF);
			break;

		case DTMF_1:
			EEP_SetCTCSSEncodeMode(MODE_ECTCSS_RUS);
			break;

		case DTMF_2:
			EEP_SetCTCSSEncodeMode(MODE_ECTCSS_PTT);
			break;

		default:
			Audio_Cancel(AU_DO_OK);		// Flag an error.
			Audio_Schedule(AU_DO_ERR);
			break;
	}
}

#endif // CTCSS_ENCODING

/****************************************************************
		End of DTMF Code Module
****************************************************************/


