/***************************************************************
*																					*
*	fgen.c: Frequency Generator											*
*																					*
*	Copyright 2006 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "bitio.h"
#include "fgen.h"

#ifdef DEBUG
#include "cdebug.h"
#include "serial.h"
#endif


	// undef define

/***************************************************************/

//		*** MACROS ****

/***************************************************************/

//		*** GENERAL DEFINITIONS ****

	//	Waveform Generator Control Words

#define FG_STOP		0b00000000	// Normal port pin functionality.

#	if PCB_VERSION == 10
#		define FG_START		0b10000011	// Fast PWM OC2A (TCCR2A).
#		define FG_PRESCALER	0b00000001	// Prescaler = 1 (TCCR2B).
#	elif PCB_VERSION == 20
#		define FG_START		0b10000011	// Fast PWM OC2A (TCCR2A).
#		define FG_PRESCALER	0b00000010	// Prescaler = 8 (TCCR2B).
#	else // PCB_VERSION == 21
#		define FG_START		0b10100011	// Fast PWM OC2A and OC2B.
#		define FG_PRESCALER	0b00000010	// Prescaler = 8 (TCCR2B).
#	endif

	// Define Phase Constant.  This is the number that the
	// desired frequency (in Hz) must be multiplied by, to obtain
	// the phase increment.  It is equal to 2^n/Rs where
	// n is the number of bits in the phase increment, and
	// Rs is the sample rate in Hz.

	// In this case, n=16 and:
	//			Rs=6.9913 KHz. (Version 1.0 PCB)
	//			Rs=9.7656 KHz. (Version 2.0 and 2.1 PCB)

	// (2^n/Rs is multiplied by 65536 to give better resolution).

	// 		K.phi = 614,330.4 (Version 1.0 PCB)
	// 		K.phi = 439,804.7 (Version 2.0 and 2.1 PCB)

	// Since we want the increment in 10ths of a Hz, we divide
	// the phase constant by 10 and round up.

#if PCB_VERSION == 10
#	define FG_FPHASE_K			61433L	// Phase constant(2^n/Rs)
#else // PCB_VERSION == 20 or 21
#	define FG_FPHASE_K			43980L	// Phase constant(2^n/Rs)
#endif

#define MAX_WAVEFORMS		2
#define MAX_FREQ				30000		// Max frequency in 1/10 Hz.

	// Gain Linearization Constants:
#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

	// Using 680 at R8, 0.47u at C27-C29 and 680 at R33:
#	define FG_TONE_SLOPE			8			// m and b for CW and CB.
#	define FG_TONE_INTERCEPT	14989

#else // PCB_VERSION == 21

	// Using 2.2K at R8, 0.1u at C27-C29 and 1.5K at R33:
#	define FG_TONE_SLOPE			6			// m and b for CW and CB.
#	define FG_TONE_INTERCEPT	26699

#endif


	// Constants for CTCSS Encoding:
//#define FG_CTCSS_SLOPE		11			// m and b for CTCSS
//#define FG_CTCSS_INTERCEPT	48889

	// Constants for CTCSS Encoding:
#define FG_CTCSS_SLOPE		5			// m and b for CTCSS (m negative)
#define FG_CTCSS_INTERCEPT	45484

	// Reverse Burst Constants

//#define FG_PHASE_REVERSE	32768		// Reverse phase 180 deg.
#define FG_PHASE_REVERSE	22756		// Reverse phase 235 deg.

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***

/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****


/***************************************************************/

//		*** VARIABLES ***

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

static byte muIsrState;

static word mwFreq0;
static byte muGain0;			// gain, 0 - 255

static word mwFreq1;
static byte muGain1;			// gain, 0 - 255

	// Waveform Parameters:

static word mwPhaseInc;
static word mwPhase;
static byte muSample;
static byte muGain;
static word mwFreq;

#else // PCB_VERSION == 21

static bool mbEnableTone;
static word mwToneFreq;
//static byte muToneGain;			// gain, 0 - 255
static byte muToneSample;
static word mwTonePhaseInc;
static word mwTonePhase;
static byte muLinearizedToneGain;

static bool mbEnableCTCSS;
static word mwCTCSSFreq;
//static byte muCTCSSGain;			// gain, 0 - 255
static byte muCTCSSSample;
static word mwCTCSSPhaseInc;
static word mwCTCSSPhase;
static byte muLinearizedCTCSSGain;

#endif

/***************************************************************/

//		*** STATIC DATA ****

	// Sine table.  Stores 1 quadrant only.

	//lint -esym(528,QuadrantTable)

static prog_uint8_t QuadrantTable[] =
{
	  2,   5,   8,  11,  14,  17,  20,  23,	// 00-07
	 26,  29,  32,  35,  38,  41,  44,  47,	// 08-0F
	 50,  53,  56,  58,  61,  64,  67,  69,	// 10-17
	 72,  74,  77,  79,  82,  84,  86,  89,	// 18-1F
	 91,  93,  95,  97,  99, 101, 103, 105,	// 20-27
	106, 108, 110, 111, 113, 114, 115, 117,	// 28-2F
	118, 119, 120, 121, 122, 123, 124, 124,	// 30-37
	125, 125, 126, 126, 127, 127, 127, 127,	// 38-3F

	127, 127, 127, 127, 126, 126, 125, 125,	// 40-47
	124, 124, 123, 122, 121, 120, 119, 118,	// 48-4F
	117, 115, 114, 113, 111, 110, 108, 106,	// 50-57
	105, 103, 101,  99,  97,  95,  93,  91,	// 58-5F
	 89,  86,  84,  82,  79,  77,  74,  72,	// 60-67
	 69,  67,  64,  61,  58,  56,  53,  50,	// 68-6F
	 47,  44,  41,  38,  35,  32,  29,  26,	// 70-77
	 23,  20,  17,  14,  11,   8,   5,   2,	// 78-7F
/*
	  2,   5,   8,  11,  14,  17,  20,  23,	// 80-87
	 26,  29,  32,  35,  38,  41,  44,  47,	// 88-8F
	 50,  53,  56,  58,  61,  64,  67,  69,	// 90-97
	 72,  74,  77,  79,  82,  84,  86,  89,	// 98-9F
	 91,  93,  95,  97,  99, 101, 103, 105,	// A0-A7
	106, 108, 110, 111, 113, 114, 115, 117,	// A8-AF
	118, 119, 120, 121, 122, 123, 124, 124,	// B0-B7
	125, 125, 126, 126, 127, 127, 127, 127,	// B8-BF

	127, 127, 127, 127, 126, 126, 125, 125,	// C0-C7
	124, 124, 123, 122, 121, 120, 119, 118,	// C8-CF
	117, 115, 114, 113, 111, 110, 108, 106,	// D0-D7
	105, 103, 101,  99,  97,  95,  93,  91,	// D8-DF
	 89,  86,  84,  82,  79,  77,  74,  72,	// E0-E7
	 69,  67,  64,  61,  58,  56,  53,  50,	// E8-EF
	 47,  44,  41,  38,  35,  32,  29,  26,	// F0-F7
	 23,  20,  17,  14,  11,   8,   5,   2		// F8-FF
*/
};


	// CTCSS Tone Frequency Table

static word __attribute__((progmem)) CTCSSEncodeTbl[] =
{
	670,	// 00:  67.0
	719,	// 01:  71.9
	744,	// 02:  74.4
	770,	// 03:  77.0
	797,	// 04:  79.7
	825,	// 05:  82.5
	854,	// 06:  85.4
	885,	// 07:  88.5
	915,	// 08:  91.5
	948,	// 09:  94.8
	1000,	// 10: 100.0
	1035,	// 11: 103.5
	1072,	// 12: 107.2
	1109,	// 13: 110.9
	1148,	// 14: 114.8
	1188,	// 15: 118.8
	1230,	// 16: 123.0
	1273,	// 17: 127.3
	1318,	// 18: 131.8
	1365,	// 19: 136.5
	1413,	// 20: 141.3
	1462,	// 21: 146.2
	1514,	// 22: 151.4
	1567,	// 23: 156.7
	1622,	// 24: 162.2
	1679,	// 25: 167.9
	1738,	// 26: 173.8
	1799,	// 27: 179.9
	1862,	// 28: 186.2
	1928,	// 29: 192.8
	2035,	// 30: 203.5
	2107,	// 31: 210.7
	2181,	// 32: 218.1
	2257,	// 33: 225.7
	2336,	// 34: 233.6
	2418,	// 35: 241.8
	2503	// 36: 250.3
};

/*****************************************************************
								Program Code
*****************************************************************/


void FGen_Init(void)
{

	TCCR2A = FG_START;		// Start waveform generation
	TCCR2B = FG_PRESCALER;

	OCR2A = 128;				// Set initial output compare.

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

	mwPhaseInc = mwPhase = 0;
	TIMSK2 &= ~BV(TOIE2);	// Disable overflow interrupts.

#else // PCB_VERSION == 21

	mbEnableTone = mbEnableCTCSS = FALSE;
	mwTonePhaseInc = mwTonePhase = 0;
	mwCTCSSPhaseInc = mwCTCSSPhase = 0;
	OCR2B = 128;				// Set initial output compare.
	TIMSK2 |= BV(TOIE2);		// Enable overflow interrupts.

#endif // PCB_VERSION

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	WfgenISR:  Two Quadrant Table


#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

SIGNAL(SIG_OVERFLOW2)
{
	byte index;
	word localSample;

	OCR2A = muSample;

	switch(muIsrState)
	{
		case 0:
			mwPhase += mwPhaseInc;		// Increment the phase.
			index = HI_BYTE(mwPhase);	// extract table index.

			localSample = pgm_read_byte((byte *)QuadrantTable
				+ (index & 0x7F)) * muGain;

			muSample = HI_BYTE(localSample);
			muSample =
				(index & 0x80)? (128-muSample) : (128+muSample);

			muIsrState = 1;

			break;

		default:
		case 1:
			muIsrState = 0;
			break;
	}
}

#else // PCB_VERSION == 21

SIGNAL(SIG_OVERFLOW2)
{
	byte index;
	word localSample;

//	DB_High();

	OCR2A = muToneSample;		// Output latest samples
	OCR2B = muCTCSSSample;

	// Calculate next tone sample:

	if(mbEnableTone)
	{
		mwTonePhase += mwTonePhaseInc;	// Increment the phase.
		index = HI_BYTE(mwTonePhase);		// extract table index.

		localSample = pgm_read_byte((byte *)QuadrantTable
			+ (index & 0x7F)) * muLinearizedToneGain;

		muToneSample = HI_BYTE(localSample);
		muToneSample =
			(index & 0x80)? (128-muToneSample) : (128+muToneSample);
	}
	else
		muToneSample = 128;

	// Calculate next CTCSS sample:

	if(mbEnableCTCSS)
	{
		mwCTCSSPhase += mwCTCSSPhaseInc;	// Increment the phase.
		index = HI_BYTE(mwCTCSSPhase);	// extract table index.

		localSample = pgm_read_byte((byte *)QuadrantTable
			+ (index & 0x7F)) * muLinearizedCTCSSGain;

		muCTCSSSample = HI_BYTE(localSample);
		muCTCSSSample =
			(index & 0x80)? (128-muCTCSSSample) : (128+muCTCSSSample);
	}
	else
		muCTCSSSample = 128;

//	DB_Low();
}

#endif // PCB_VERSION


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Start waveform generation.

#if((PCB_VERSION == 10) || (PCB_VERSION == 20))

void FGen_Start(byte waveNum)
{
	byte ieStatus;

	switch(waveNum)
	{
		default:
		case FGW_TONE:
			FGen_SetLinearizedToneGain();
			mwFreq = mwFreq0;			// select frequency.

#	if(PCB_VERSION == 10)
			IO_Set(O_CWIDENA);		// Disable attenuation.
#	endif // PCB_VERSION

			break;

		case FGW_CTCSS:
			mwFreq = mwFreq1;			// select frequency.
			muGain = muGain1;			// select gain.

#	if(PCB_VERSION == 10)
			IO_Clear(O_CWIDENA);		// Enable attenuation.
#	endif // PCB_VERSION

			break;
	}

	// Init phase counters

	mwPhaseInc = (((mwFreq * FG_FPHASE_K) + 0x8000L) >> 16);
	mwPhase = 0;		// Start at a zero crossing.

	DI(ieStatus);		// Nested disable interrupts

	OCR2A = 128;			// Set initial output compare.

	// Enable phase counter interrupts

	TIMSK2 |= BV(TOIE2);		// Enable overflow interrupts.

	EI(ieStatus);		// Nested enable interrupts
}

#else // PCB_VERSION == 21

void FGen_Start(byte waveNum)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	switch(waveNum)
	{
		default:
		case FGW_TONE:
			mbEnableTone = TRUE;
			break;

		case FGW_CTCSS:
			mbEnableCTCSS = TRUE;
			break;
	}

	EI(ieStatus);		// Nested enable interrupts

}

#endif // PCB_VERSION

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Stop waveform generation by disabling interrupts.

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

void FGen_Stop(byte waveNum)
{
	byte ieStatus;
	DI(ieStatus);		// Nested disable interrupts

	TIMSK2 &= ~BV(TOIE2);	// Disable overflow interrupts.
	OCR2A = 128;				// Set initial output compare.

	mwPhase = mwPhaseInc = 0;

#	if(PCB_VERSION == 10)
	IO_Clear(O_CWIDENA);		// Enable attenuation.
#	endif

	EI(ieStatus);		// Nested enable interrupts
}

#else // PCB_VERSION == 21

void FGen_Stop(byte waveNum)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	switch(waveNum)
	{
		default:
		case FGW_TONE:
			mbEnableTone = FALSE;
			break;

		case FGW_CTCSS:
			mbEnableCTCSS = FALSE;
			break;
	}

	EI(ieStatus);		// Nested enable interrupts
}

#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#ifdef CTCSS_ENCODING

void FGen_EnableCTCSSGeneration(byte code)
{
	byte *pFreq;
	byte uFreqL, uFreqH;
	word wFreq;

//Serial_TxChar('1');
	pFreq = (byte *)CTCSSEncodeTbl + (code * sizeof(word));

	uFreqL = PRG_RDB(pFreq++);
	uFreqH = PRG_RDB(pFreq);
	wFreq = MAKEWORD(uFreqH,uFreqL);

	FGen_SetFreq(FGW_CTCSS, wFreq);
	FGen_Start(FGW_CTCSS);
}

void FGen_DisableCTCSSGeneration(void)
{
//Serial_TxChar('0');

	FGen_Stop(FGW_CTCSS);
}

	// Reverse the phase by 120 degrees.

void FGen_CTCSSReverseBurst(void)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts
	mwCTCSSPhase += FG_PHASE_REVERSE;
	EI(ieStatus);		// Nested enable interrupts
}

#endif // CTCSS_ENCODING

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set the frequency of a waveform.

	// waveNum	0 or 1
	// freq is in 10ths of a Hz.

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

void FGen_SetFreq(byte waveNum, word freq)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	if(waveNum < MAX_WAVEFORMS)
	{
		switch(waveNum)
		{
			default:
			case FGW_TONE:
				mwFreq0 = freq;
				FGen_SetLinearizedToneGain();
				break;

			case FGW_CTCSS:
				mwFreq1 = freq;
				break;
		}
	}

	EI(ieStatus);		// Nested enable interrupts
}

#else // PCB_VERSION == 21

void FGen_SetFreq(byte waveNum, word freq)
{
	byte ieStatus;
	word phaseInc;
	word temp1;

	phaseInc = (((freq * FG_FPHASE_K) + 0x8000L) >> 16);

	DI(ieStatus);		// Nested disable interrupts

	if(waveNum < MAX_WAVEFORMS)
	{
		switch(waveNum)
		{
			default:
			case FGW_TONE:
				mwToneFreq = freq;
				mwTonePhaseInc = phaseInc;
				mwTonePhase = 0;		// Start at a zero crossing.
				break;

			case FGW_CTCSS:
				mwCTCSSFreq = freq;
				mwCTCSSPhaseInc = phaseInc;
				mwCTCSSPhase = 0;		// Start at a zero crossing.

				// We're not even looking at the counts here.
				// We just set the level based on the frequency.

				temp1 = ((word)FG_CTCSS_INTERCEPT -
					(mwCTCSSFreq * (byte)FG_CTCSS_SLOPE)) + 128;

				DI(ieStatus);		// Nested disable interrupts
				muLinearizedCTCSSGain = HI_BYTE(temp1);
				EI(ieStatus);		// Nested enable interrupts
				break;
		}
	}

	EI(ieStatus);		// Nested enable interrupts
}


#endif // PCB_VERSION

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Set the gain of the waveform (in percent).
	// level ranges from 0 100, but gain is 0 to 255
	// so we must scale level.

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

void FGen_SetGain(byte waveNum, byte level)
{
	byte ieStatus;
	byte counts;

	DI(ieStatus);		// Nested disable interrupts

	if(waveNum < MAX_WAVEFORMS)
	{
		counts = HI_BYTE((level * 654) + 128);

		switch(waveNum)
		{
			default:
			case FGW_TONE:
				muGain0 = counts;
				FGen_SetLinearizedToneGain();
				break;

			// Set up for non-linearization.  Range is small
			// enough so that linearization is not required.

			case FGW_CTCSS:
				muGain1 = counts;
				break;
		}
	}

	EI(ieStatus);		// Nested enable interrupts
}

#else // PCB_VERSION == 21

void FGen_SetGain(byte waveNum, byte level)
{
	byte ieStatus;
	byte counts;
	word temp1;


	if(waveNum < MAX_WAVEFORMS)
	{
		counts = HI_BYTE((level * 654) + 128);

		switch(waveNum)
		{
			default:
			case FGW_TONE:
				temp1 = HI_BYTE(((mwToneFreq - 6000) * FG_TONE_SLOPE)
					+ FG_TONE_INTERCEPT + 128);
				// Round up and divide by 256

				DI(ieStatus);		// Nested disable interrupts
//				muLinearizedToneGain = counts;
				muLinearizedToneGain = HI_BYTE(((temp1*counts) + 128));
				EI(ieStatus);		// Nested enable interrupts
				break;

			// This case is used for debug purposes only (G cmd).
			// The level for CTCSS encoding is based on the
			// frequency only.

			case FGW_CTCSS:
				DI(ieStatus);		// Nested disable interrupts
				muLinearizedCTCSSGain = counts;
				EI(ieStatus);		// Nested enable interrupts
				break;
		}
	}
}

#endif // PCB_VERSION

#if ((PCB_VERSION == 10) || (PCB_VERSION == 20))

void FGen_SetLinearizedToneGain(void)
{
	word temp1;

	temp1 = ((mwFreq0 - 6000) * FG_TONE_SLOPE) + FG_TONE_INTERCEPT;
	temp1 = HI_BYTE((temp1 + 128));	// Round up and divide by 256
	muGain = HI_BYTE(((temp1*muGain0) + 128));
}

#endif // PCB_VERSION

/*
void FGen_SetLinearizedCTCSSGain(void)
{
	muGain1 = ((word)FG_INTERCEPT1 -
		(mwFreq1 * (byte)FG_SLOPE1)) >> 8;
}

*/

/****************************************************************
					End of Waveform Generator Module
****************************************************************/
