/***************************************************************
*																					*
*	plsm.c: State Machine for detecting PL Tones						*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*																					*
*	Copyright 2004 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <limits.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "serial.h"
#include "plsm.h"
#include "parse.h"
#include "fgen.h"
#include "bitio.h"

#ifdef DEBUG
#include "cdebug.h"
#endif


	// undef define

#ifdef PL_DEBUG
#define PL_DEBUG_STATE_FLAGS	// Show state flags on monitor
#undef PL_DEBUG_EV_DISP			// Show event messages
#undef PL_DEBUG_EDGE_NUMS		// Show raw edge data
#undef PL_DEBUG_STATS			// Show statistical data
#endif



/***************************************************************/

//		*** MACROS ****

#if defined PL_DEBUG_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y)|0x20)
#else
#define SET_STATE(X,Y)	Ex_SetState(X)
#endif


#define ABS(X)			(((X) > 0)? (X) : -(X))

/***************************************************************/

//		*** GENERAL DEFINITIONS ****

	// Sample buffer size:

#define L2_Z_SAMPLE_BUF	3		// Log2 of sample buf size
#define Z_SAMPLE_BUF		(1 << L2_Z_SAMPLE_BUF)

#define T_PL_SQUELCH_TO	250	// mS from squelch to PL reset.
//#define PL_TIMEOUT		1000	// Number of ms for squelch timeout.

	// Reverse Burst Constants

//#define PL_PHASE_REVERSE	32768		// Reverse phase 180 deg.
#define PL_PHASE_REVERSE	22756		// Reverse phase 235 deg.

/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void PLTmrStart(void);

static void Transition_State_A(void);
static void State_A(MSG *pEvent);
static void Transition_State_B(void);
static void State_B(MSG *pEvent);
//static void Transition_State_C(void);
//static void State_C(MSG *pEvent);
static void Transition_State_D(void);
static void State_D(MSG *pEvent);

static void InitPLFilter(void);
static void ProcessSample(word sample);

static void DisableSampler(void);
static void EnableSampler(void);
static void LockSampler(void);

/***************************************************************/

//		*** VARIABLES ***

static struct TAG_CTCSS_STATS
{
	word idealPeriod;		// Ideal period of tone.
	word shortPeriod;		// If sample is shorter than this...
	word longPeriod;		// .. or longer than this, it's not real.

	int captTolerance;	// Capture tolerance.
	int holdTolerance;	// Hold tolerance.
	word devCaptTolerance;	// Deviation capture tolerance.
	word devHoldTolerance;	// Deviation hold tolerance.
	int avgError;			// Average error.

} plStats;


	// Interrupt Related Variables:

static byte ic1State;		// State of input capture interrupt.
static word shortSample;	// Sample too short to be real.
static word longSample;		// Sample too long to be real.


	// Sample Filter Statistics

static int sampleBuf[Z_SAMPLE_BUF];	// Sample averaging buffer
static byte sampleCount;	// Samples left before filter settled.
static word avgDeviation;	// Average deviation from the mean.

static byte lockChannel;	// Number of current channel in lock.
static bool locked;			// TRUE if PL Decoder locked on freq.


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	PL Ideal Period Table (Resides in Flash)

static word __attribute__((progmem)) PLCodeTbl[] =
{
#if PCB_VERSION == 10		// Clock 3.579 MHz, Prescaler = 8

	6678,	// 00:  67.0
	6223,	// 01:  71.9
	6014,	// 02:  74.4
	5811,	// 03:  77.0
	5614,	// 04:  79.7
	5424,	// 05:  82.5
	5239,	// 06:  85.4
	5056,	// 07:  88.5
	4890,	// 08:  91.5
	4720,	// 09:  94.8
	4474,	// 10: 100.0
	4323,	// 11: 103.5
	4174,	// 12: 107.2
	4035,	// 13: 110.9
	3898,	// 14: 114.8
	3766,	// 15: 118.8
	3638,	// 16: 123.0
	3515,	// 17: 127.3
	3395,	// 18: 131.8
	3278,	// 19: 136.5
	3167,	// 20: 141.3
	3060,	// 21: 146.2
	2955,	// 22: 151.4
	2855,	// 23: 156.7
	2759,	// 24: 162.2
	2665,	// 25: 167.9
	2574,	// 26: 173.8
	2487,	// 27: 179.9
	2403,	// 28: 186.2
	2321,	// 29: 192.8
	2199,	// 30: 203.5
	2124,	// 31: 210.7
	2052,	// 32: 218.1
	1982,	// 33: 225.7
	1915,	// 34: 233.6
	1850,	// 35: 241.8
	1788	// 36: 250.3

#define SHORT_SAMPLE		1609		// 10% shorter than tone 36
#define LONG_SAMPLE		7346		// 10% longer than tone 0

#else	// PCB_VERSION == 20 or 21	(Clock 20.0 MHz, Prescaler = 64)

	4664,	// 00: 67.0
	4346,	// 01: 71.9
	4200,	// 02: 74.4
	4058,	// 03: 77.0
	3921,	// 04: 79.7
	3788,	// 05: 82.5
	3659,	// 06: 85.4
	3531,	// 07: 88.5
	3415,	// 08: 91.5
	3296,	// 09: 94.8
	3125,	// 10: 100.0
	3019,	// 11: 103.5
	2915,	// 12: 107.2
	2818,	// 13: 110.9
	2722,	// 14: 114.8
	2630,	// 15: 118.8
	2541,	// 16: 123.0
	2455,	// 17: 127.3
	2371,	// 18: 131.8
	2289,	// 19: 136.5
	2212,	// 20: 141.3
	2137,	// 21: 146.2
	2064,	// 22: 151.4
	1994,	// 23: 156.7
	1927,	// 24: 162.2
	1861,	// 25: 167.9
	1798,	// 26: 173.8
	1737,	// 27: 179.9
	1678,	// 28: 186.2
	1621,	// 29: 192.8
	1536,	// 30: 203.5
	1483,	// 31: 210.7
	1433,	// 32: 218.1
	1385,	// 33: 225.7
	1338,	// 34: 233.6
	1292,	// 35: 241.8
	1249	// 36: 250.3

#define SHORT_SAMPLE		1124		// 10% shorter than tone 36
#define LONG_SAMPLE		5130		// 10% longer than tone 0

#endif // PCB_VERSION



	// If you add more here, be sure to change the MAX_PL_CODE
	// constant in plsm.h
};

/***************************************************************/

//		*** API ***

void PL_Init(void)
{
	// Init Timer 1 for Input Capture:

	TCCR1A = 0x00;			// No output compare modes enabled

	TCCR1B = 0x43;			// Noise Canceler off.
								// Capture rising edge. Div by 64 PS

	lockChannel = 0;		// No channel in lock.
	locked = FALSE;

	PL_SetToneDecodeThresholds();
	Transition_State_A();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	// Here, return TRUE if we are locked on a PL tone,
	// or if no tone is required.

bool PL_IsUnsquelched(void)
{
	if(EEP_DecodePL())
		return(locked);

	else
		return(TRUE);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	Turn on the PL Tone detector

	//	Select the PL Tone Paramaters

void PL_SetToneDecodeThresholds(void)
{
	byte *pPeriod;
	byte periodL, periodH;
	word idealPeriod;

#ifdef PL_TWEEK
	byte captTol, captDev, holdTol, holdDev;
#endif

	// Read the table and get the ideal period count.
	pPeriod = (byte *)PLCodeTbl +
		(EEP_GetPLDecodeNum() * sizeof(word));

	periodL = PRG_RDB(pPeriod++);
	periodH = PRG_RDB(pPeriod);
	idealPeriod = MAKEWORD(periodH,periodL);
	plStats.idealPeriod = idealPeriod;

	// Compute the capture and hold tolerances.  If the
	// average error is less than the capture tolerance,
	// we can go into lock.  If in lock and the average
	// error is less than the hold tolerance, we can stay
	// in lock.  If the deviation is greater than the
	// devTolerance we go out of lock.

#ifdef PL_TWEEK
	captTol = 8;
	holdTol = 4; holdDev = 3;

	switch(EEP_GetPLTweek())
	{
		default:
		case 0: captDev = 5; holdDev = 3; break;
		case 1: captDev = 4; holdDev = 3; break;
		case 2: captDev = 3; holdDev = 3; break;
		case 3: captDev = 2; holdDev = 3; break;

		case 4: captDev = 2; holdDev = 3; break;
		case 5: captDev = 2; holdDev = 2; break;
	}

	plStats.captTolerance = (idealPeriod >> captTol);
	plStats.devCaptTolerance = (idealPeriod >> captDev);

	plStats.holdTolerance = (idealPeriod >> holdTol);
	plStats.devHoldTolerance = (idealPeriod >> holdDev);

	Serial_CRLF();
	Serial_TxChar(captTol + '0');
	Serial_TxChar(captDev + '0');
	Serial_TxChar(holdTol + '0');
	Serial_TxChar(holdDev + '0');

#else
	// 090201 Changing captTolerance from 8:
	plStats.captTolerance = (idealPeriod >> 6);

	plStats.devCaptTolerance = (idealPeriod >> 5);

	plStats.holdTolerance = (idealPeriod >> 4);
	plStats.devHoldTolerance = (idealPeriod >> 3);

#endif

	plStats.shortPeriod = idealPeriod - (idealPeriod >> 2);
	plStats.longPeriod = idealPeriod + (idealPeriod >> 2);

#ifdef PL_SIMULATION
		ReportThresholds(idealPeriod,
			plStats.captTolerance,
			plStats.devCaptTolerance,
			plStats.holdTolerance,
			plStats.devHoldTolerance);
#endif

	InitPLFilter();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	PL Tone Interrupt Handlers

//	Timer 1 Input Capture Interrupt Handler
//	This interrupt occurs when a new PL tone edge is detected.

SIGNAL(SIG_INPUT_CAPTURE1)
{
	static word prevEdge;	// Previous edge time.
	word rawEdge;				// Raw edge time
	word rawSample;
	byte tempL, tempH;

	// Read input capture reg.
	tempL = inp(ICR1L); tempH = inp(ICR1H);
	rawEdge = MAKEWORD(tempH,tempL);

#ifdef PL_DEBUG_EDGE_NUMS
	DB_WORD(rawEdge);
	DB_CRLF();
#endif

	switch(ic1State)
	{
		// Get a starting edge:

		default:
		case 0:
			prevEdge = rawEdge;
			ic1State = 1;
			break;

		// Locking - Get a new sample:

		case 1:
			rawSample = rawEdge - prevEdge;
			prevEdge = rawEdge;

			// Tell PL detector state machine that a new edge occured.

			if((rawSample >= SHORT_SAMPLE) &&
				(rawSample <= LONG_SAMPLE))
				Ex_SendEventExt(SM_PL, EV_PL_SAMPLE, rawSample);
			else
				ic1State = 0;

			break;

		// Locked - Get a new sample.  If the sample is too short,
		// wait for the next edge.

		case 2:
			rawSample = rawEdge - prevEdge;

			// Check raw sample to make sure it is in range.

			if(rawSample >= shortSample)
			{
				if(rawSample <= longSample)
				{
					Ex_SendEventExt(SM_PL, EV_PL_SAMPLE, rawSample);
				}

				prevEdge = rawEdge;
			}
			break;
	}

	// Update time limit for missed edge.

	rawEdge += LONG_SAMPLE;
	OCR1AH = HI_BYTE(rawEdge);
	OCR1AL = LO_BYTE(rawEdge);

	// Clear latent OC1A irq
	TIFR1 = BV(OCF1A);

	// Enable OC1A irqs.
	TIMSK1 |= BV(OCIE1A);
}


	//	Timer 1 Output Compare A Handler:
	//	This interrupt occurs when an expected PL Tone edge
	//	takes too long to appear.

SIGNAL(SIG_OUTPUT_COMPARE1A)
{
	ic1State = 0;		// Force a new starting edge.

	// Disable further OC1A irqs. (Wait for next IC1).
	TIMSK1 &= ~(BV(OCIE1A));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	// Disable and enable the Tone Sampling System

static void DisableSampler(void)
{
	// Disable further IC1 and OC1A irqs.
	TIMSK1 &= ~(BV(ICIE1) | BV(OCIE1A));
	ic1State = 0;		// Force a new starting edge.
}

static void EnableSampler(void)
{
	ic1State = 0;		// Force a new starting edge.

	// Clear latent IC1 irqs
	TIFR1 = BV(ICF1);

	// Enable IC1 irq.
	TIMSK1 |= BV(ICIE1);
}

static void LockSampler(void)
{
	// Disable IC1 irqs.
	TIMSK1 &= ~BV(ICIE1);

	shortSample = plStats.shortPeriod;
	longSample = plStats.longPeriod;

	ic1State = 2;		// Start using tighter limits.

	// Enable IC1 irq.
	TIMSK1 |= BV(ICIE1);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void PLTmrStart(void)
{
//	Ex_StartEventTimer(TM_PL_TIMEOUT, TP_1MS,
//		PL_TIMEOUT, EV_PL_TIMEOUT);

	Ex_StartEventTimer(TM_PL_TIMEOUT, TP_1MS,
		EEP_GetPLDecodeTimeout(), EV_PL_TIMEOUT);
}

#define PLTmrStop()	Ex_StopTimer(TM_PL_TIMEOUT)


static void SquelchTmrStart(void)
{
	Ex_StartEventTimer(TM_PL_SQUELCH, TP_1MS,
		T_PL_SQUELCH_TO, EV_PL_SQUELCH_TO);
}

#define SquelchTmrStop()	Ex_StopTimer(TM_PL_SQUELCH)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	Initialize the state machine to state A:

static void Transition_State_A(void)
{
	locked = FALSE;

	DisableSampler();
	InitPLFilter();
	Ex_SendEvent(SM_RX, EV_PL_UNLOCKED);	// Say not locked.
//DB_Low();
	PLTmrStop();			// Stop lock timer
	SquelchTmrStop();		// Stop squelch timer

	// Select next state = A
	SET_STATE(State_A, 'A');
}


	//	PLStateA: No PL has been detected.  Check to see if we are
	//	receiving a new edge.

static void State_A(MSG *pEvent)
{
	switch(pEvent->id)
	{
		case MSG_PL_UNSQUELCHED:
			if(EEP_DecodePL())
			{
				EnableSampler();	// Start the interrupts.
				Transition_State_B();
			}
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	PL State B: Unlocked.  An initial edge was received.  Process
//	samples and compare to capture criteria.

static void Transition_State_B(void)
{
	InitPLFilter();
	locked = FALSE;

	// Select next state = B
	SET_STATE(State_B, 'B');
}

static void State_B(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// Receiver is squelched, but squelch may be bouncing.

		case MSG_PL_SQUELCHED:
			SquelchTmrStart();
			break;

		// Here, receiver went unsquelched.

		case MSG_PL_UNSQUELCHED:
			SquelchTmrStop();
			break;

		// Squelch timer timed out.  Squelch is debounced.

		case EV_PL_SQUELCH_TO:
			Transition_State_A();
			break;

		// Here we have received a new edge.

		case EV_PL_SAMPLE:
			ProcessSample(pEvent->wData);

			// Here, test to see if the average error is less
			// than the capture criterion.

			if((plStats.avgError < plStats.captTolerance)
				&& (avgDeviation < plStats.devCaptTolerance))
			{
				// Here, we are locked.
				Ex_SendEvent(SM_RX, EV_PL_LOCKED);
				Transition_State_D();
			}

			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

	//	PL State D: Locked.  Process samples and compare to
	//	hold criteria.

static void Transition_State_D(void)
{
	locked = TRUE;
	LockSampler();
//DB_High();
	PLTmrStart();	// Start lock timer

	// Select next state = D
	SET_STATE(State_D, 'D');
}


static void State_D(MSG *pEvent)
{
	switch(pEvent->id)
	{
		// Receiver is squelched, but squelch may be bouncing.

		case MSG_PL_SQUELCHED:
			SquelchTmrStart();
			break;

		// Here, receiver went unsquelched.

		case MSG_PL_UNSQUELCHED:
			SquelchTmrStop();
			break;

		// Squelch timer timed out.  Squelch is debounced.

		case EV_PL_SQUELCH_TO:
			Transition_State_A();
			break;

		// Here, the lock timer timed out.  PL is unlocked.

		case EV_PL_TIMEOUT:
//DB_Low();

			Ex_SendEvent(SM_RX, EV_PL_UNLOCKED);
			Transition_State_B();
			break;

		// Here we have received a new edge.

		case EV_PL_SAMPLE:
			ProcessSample(pEvent->wData);

			// Here, test to see if the average error is
			// less than the hold criterion.

			if((plStats.avgError
				< plStats.holdTolerance) &&
				(avgDeviation < plStats.devHoldTolerance))
			{
//DB_High();
				PLTmrStart();	// Refresh lock timer
			}
			break;

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void InitPLFilter(void)
{
	sampleCount = Z_SAMPLE_BUF;
}

static void ProcessSample(word sample)
{
	byte i;
	word tempSample, avgSamples;
	int deviation;
	word sumSamples;
	word sumDeviation;
	int avgError;

	sumSamples = sample;

	// Shift and accumulate the error buffer.

	for(i=Z_SAMPLE_BUF-1; i>0; i--)
	{
		tempSample = sampleBuf[i-1];	// Shift
		sampleBuf[i] = tempSample;
		sumSamples += tempSample;
	}

	sampleBuf[0] = sample;	// Put newest sample in buffer

	// If the buffer is not full, wait.

	if(sampleCount > 0)
	{
		sampleCount--;
		avgDeviation = avgSamples = sumSamples = UINT_MAX;

	}

	// Here, buffer is full.  Calculate statistics.

	else
	{
		avgSamples = sumSamples >> L2_Z_SAMPLE_BUF;	// Get mean.

		// Compute the deviation from the average

		i=Z_SAMPLE_BUF-1;
		sumDeviation = 0;

		do
		{
			deviation = sampleBuf[i--] - avgSamples;
			sumDeviation += ABS(deviation);
		}
		while(i>0);

		avgDeviation = sumDeviation >> L2_Z_SAMPLE_BUF;

	}

	avgError = avgSamples - plStats.idealPeriod;
	plStats.avgError = ABS(avgError);

#ifdef PL_DEBUG_STATS
//			DB_WORD(sample);
//			DB_WORD(avgSamples);
//			DB_WORD(avgDeviation);
//			DB_CRLF();
#endif

#ifdef PL_SIMULATION
	ReportStatistics(avgSamples, avgDeviation);
#endif

}

/****************************************************************
		End of PL Tone Detection Module
****************************************************************/
