/***************************************************************
*																					*
*	rxsm.c: Receiver State Machine										*
*																					*
*	Originally developed for PSE-407 Repeater Controller			*
*	Modified for single channel use in PSE-508 controller.		*
*																					*
*	Copyright 2006 by Pion and Simon Electronics.  All				*
*	rights reserved.															*
*																					*
    This file is part of the PSE-508 project

    The PSE-508 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-508 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

//		*** INCLUDE FILES ***

#include <avr/io.h>
#include <avr/interrupt.h>

#include "avrstd.h"
#include "app.h"
#include "avrapex.h"
#include "bitio.h"
#include "rxsm.h"
#include "plsm.h"
#include "dtmf.h"
#include "parse.h"
#include "serial.h"

#ifdef DEBUG
#include "cdebug.h"
#endif


	// undef define

#ifdef RX_DEBUG
#undef RX_DEBUG_UC_STATE_FLAGS	// Show UC state flags on monitor
#define RX_DEBUG_LC_STATE_FLAGS	// Show LC state flags on monitor
#define RX_DEBUG_EV_DISP		// Show event messages
#endif

/***************************************************************/

//		*** MACROS ****

#if defined RX_DEBUG_UC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y))
#elif defined RX_DEBUG_LC_STATE_FLAGS
#define SET_STATE(X,Y)	DB_SetState((X), (Y)|0x20)
#else
#define SET_STATE(X,Y)	Ex_SetState(X)
#endif


/***************************************************************/

//		*** GENERAL DEFINITIONS ****

#define T_VID_REC_DLY	10000		// Voice ID Record Delay, mS
#define T_VID_REC_DUR	10000		// Voice ID Max Duration in mS
#define T_SQUELCH_TAIL	200		// Voice ID tail cutoff in mS
#define T_SQUELCH_POLL	10			// mS between squelch polls.


	// Definition of Flags in Flag Byte

//#define F_				0x01		// Unused.
//#define F_				0x02		// Unused.
//#define F_				0x04		// Unused.
//#define F_				0x08		// Unused.
#define F_SQU_PREV		0x10		// Saved squelch state from prev. poll.
#define F_RX_UNMUTED		0x20		// Set if channel's Rx is unmuted.
//#define F_TX_ID			0x40		// Set if we need to ID on chan
//#define F_				0x80		// Unused.



/***************************************************************/

//		*** DATA STRUCTURE Definitions ***


/***************************************************************/

//		*** STATIC FUNCTION PROTOTYPES ****

static void MuteChanRx(void);
static void UnmuteChanRx(void);
static void PollSquelchISR(void);

static void Transition_State_A(void);
static void State_A(MSG *pEvent);

#ifdef VOICE_ID
static void Transition_State_B(void);
static void State_B(MSG *pEvent);
#endif // VOICE_ID

static void Transition_State_C(void);
static void State_C(MSG *pEvent);
static void Transition_State_D(void);
static void State_D(MSG *pEvent);

#ifdef VOICE_ID
static void Transition_State_E(void);
static void State_E(MSG *pEvent);
static void Transition_State_F(void);
static void State_F(MSG *pEvent);
#endif // VOICE_ID


/***************************************************************/

//		*** VARIABLES ***

#ifdef VOICE_ID
static unsigned long vidStartTime;	// System time when recording started.
#endif // VOICE_ID

static bool unsquelched;	// Receive is unsquelched.
static byte flags;			// Chan status flags

/*****************************************************************

								Global Functions

	The following functions provide the external interface (API)
	for the Receiver State Machine.

*****************************************************************/


void Rx_Init(void)
{
	unsquelched = FALSE;
	flags = F_SQU_PREV;
	Ex_StartPollTimer(
		TM_RX_SQUELCH_POLL, T_SQUELCH_POLL, PollSquelchISR);
	Transition_State_A();		// Select state A
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Poll the squelch status.

// This function is called from a timer interrupt service
// routine every 10 mS.

// ALP 060525: Checked to see if interrupts should be disabled
//					around this routine.  It appears that they
//					do not need to be.

static void PollSquelchISR(void)
{
	if(IO_Get(I_SQUELCH))		// active high!
	{
		// Here, channel is squelched.

		if(!(flags & F_SQU_PREV)) // was it unsquelched before?
		{
			// Here, channel went from unsquelched to squelched.

			flags |= F_SQU_PREV; // Flag it squelched.
			Ex_SendEvent(SM_RX, EV_SQUELCHED);
			Ex_SendMsg(SM_PL, MSG_PL_SQUELCHED);
		}
	}

	// Here, channel is unsquelch.

	else
	{

		if(flags & F_SQU_PREV) // was chan squelched before?
		{
			// Here, channel went from squelched to unsquelched.

			flags &= ~(F_SQU_PREV);	// Flag it unsquelched.
			Ex_SendEvent(SM_RX, EV_UNSQUELCHED);
			Ex_SendMsg(SM_PL, MSG_PL_UNSQUELCHED);
		}
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void UnmuteChanRx(void)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	IO_Set(O_RXENA);		// Unmute the chan.
	flags |= F_RX_UNMUTED;
	IO_Set(O_UNSQULED);	// Turn on the LED.

	EI(ieStatus);		// Nested enable interrupts
}

static void MuteChanRx(void)
{
	byte ieStatus;

	DI(ieStatus);		// Nested disable interrupts

	IO_Clear(O_RXENA);		// Mute the chan.
	flags &= ~F_RX_UNMUTED;
	IO_Clear(O_UNSQULED);	// Turn off the LED.

	EI(ieStatus);		// Nested enable interrupts
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//	Execute the Receiver State Machine.

void Rx_SM(MSG *pEvent)
{
#ifdef RX_DEBUG_EV_DISP
DB_PrintEvent(pEvent);
#endif

	switch(pEvent->id)
	{
		// Squelch Opened

		case EV_UNSQUELCHED:
			unsquelched = TRUE;
			break;

		case EV_SQUELCHED:
			unsquelched = FALSE;
			break;

		default:
			break;
	}
}


/*****************************************************************

								Static Functions

	The following functions provide the internal functionality
	of the Receiver State Machine and are not accessable by
	external modules.

*****************************************************************/

// Receiver State A:  Wait for receive channel to go unsquelched.

static void Transition_State_A(void)
{
	Ex_SendMsg(SM_AUDIO, MSG_AUD_SQ);	// Disable PL out
	MuteChanRx();			// Turn the channel off.
	SET_STATE(State_A, 'A');
}

static void State_A(MSG *pEvent)
{
	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Squelch Opened

		case EV_UNSQUELCHED:
		case EV_PL_LOCKED:
		{
			if(unsquelched && PL_IsUnsquelched())
			{
				// Here, the channel has been enabled by the
				// user.

				UnmuteChanRx();		// Unmute the channel's receiver.

				if(EEP_TxIsEnabled())		// Is it Rx only?
					Transition_State_C();	// If not, repeating.
				else
					Transition_State_D();	// Otherwise don't enable Tx.
			}
		}
		break;

#ifdef VOICE_ID
		// Voice ID Recording is being Enabled via Serial I/O.

		case MSG_VID_ENA:
			Transition_State_B();	// Recording voice ID.
			break;

#endif

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Receiver State B:  Recording Voice ID.

// Come here if a command was received to prepare to record
// the voice ID.  Start a timer and
// if recording does not begin before the timer times out,
// cancel the recording.  Stay here until either squelch is
// opened (in which case, commence recording), or the timer
// times out (in which case, cancel recording).

#ifdef VOICE_ID

static void Transition_State_B(void)
{
	Ex_StartEventTimer(
		TM_RX_VID_TIMEOUT, TP_1MS, T_VID_REC_DLY, EV_VID_TO);

	Ex_SetState(State_B);
	SET_STATE(State_B, 'B');
}

static void State_B(MSG *pEvent)
{
	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Squelch Opened

		case EV_UNSQUELCHED:
		case EV_PL_LOCKED:
			if(unsquelched && PL_IsUnsquelched())
			{
				// Here, the channel has been enabled by the
				// user and is unsquelched.  Begin recording voice ID.

				Ex_StopTimer(TM_RX_VID_TIMEOUT);
				UnmuteChanRx();		// Unmute the channel's receiver.
				IO_Set(O_VIDREC);		// Start the record.
				vidStartTime = Ex_GetSystemTime();	// Get time it started.
				Transition_State_E();				// Wait for user to end it.
			}
			break;

		case EV_SQUELCHED:
		case EV_PL_UNLOCKED:
			DTMF_Disable();			// Disable DTMF detection
			Ex_SendMsg(SM_TX, MSG_TX_ENDRPT);
			MuteChanRx();			// Turn the channel off.
			break;

		// Here, the user didn't actually record anything before
		// the voice ID timer timed out.

		case EV_VID_TO:
			Transition_State_A();
			break;

		default:
			break;
	}
}

#endif // VOICE_ID

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Receiver State C:  Repeating

static void Transition_State_C(void)
{
	Ex_SendMsg(SM_AUDIO, MSG_AUD_UNSQ);	// Enable PL out

	DTMF_Enable();		// Enable DTMF detection

	// Enable the transmitters.
	Ex_SendMsg(SM_TX, MSG_TX_REPEAT);

	// Select next state = C
	SET_STATE(State_C, 'C');
}

static void State_C(MSG *pEvent)
{
	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Here, the COR went inactive.  Done repeating.

		case EV_SQUELCHED:
		case EV_PL_UNLOCKED:
			// Turn transmitter off.
			if(!unsquelched || !PL_IsUnsquelched())
			{
				DTMF_Disable();			// Disable DTMF detection
				Ex_SendMsg(SM_TX, MSG_TX_ENDRPT);
				Transition_State_A();	// Wait for next COR.
			}
			break;

		// Here, Tx timed out.  Mute the Rx channel and stay
		// in this state to wait for COR to drop.

		case EV_RX_TXTO:
			DTMF_Disable();		// Disable DTMF detection
			MuteChanRx();			// Turn the Rx channel off.
			Ex_SendMsg(SM_AUDIO, MSG_AUD_SQ);	// Disable PL out
			break;

#ifdef VOICE_ID

		// Voice ID Recording is being Enabled.

		case MSG_VID_ENA:
			Transition_State_F();	// Wait for unkey
			break;

#endif // VOICE_ID

		default:
			break;
	}
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Receiver State D:  Receive only.  Enable DTMF but do not
// enable Tx.

static void Transition_State_D(void)
{
	DTMF_Enable();		// Enable DTMF detection

	// Select next state = D
	SET_STATE(State_D, 'D');
}

static void State_D(MSG *pEvent)
{
	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Here, the COR went inactive.  Return to A.

		case EV_SQUELCHED:
		case EV_PL_UNLOCKED:
			DTMF_Disable();			// Disable DTMF detection

			if(EEP_TxIsEnabled())	// Did Tx get turned back on?
				Ex_SendMsg(SM_TX, MSG_TX_ON);

			Transition_State_A();	// Wait for next COR.
			break;

		// Voice ID Recording is being Enabled.

#ifdef VOICE_ID

		case MSG_VID_ENA:
			Transition_State_F();	// Wait for unkey.
			break;

#endif // VOICE_ID

		default:
			break;
	}
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Receiver State E: Record Voice ID

#ifdef VOICE_ID

static void Transition_State_E(void)
{
	// Select next state = E
	SET_STATE(State_E, 'E');
}

static void State_E(MSG *pEvent)
{
	unsigned long vidDuration;

	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Here, the COR went inactive.  Since we were recording
		// a voice ID, we need to end it.

		case EV_SQUELCHED:
		case EV_PL_UNLOCKED:
			IO_Clear(O_VIDREC);
			vidDuration =
				Ex_GetSystemTime() - vidStartTime - T_SQUELCH_TAIL;

			vidDuration =
				vidDuration > (T_VID_REC_DUR - T_SQUELCH_TAIL)?
				(unsigned long)(T_VID_REC_DUR - T_SQUELCH_TAIL)
				: vidDuration;

			EEP_SetVoiceIDDuration((word)vidDuration);
			Transition_State_A();	// Wait for next COR.
			break;

		default:
			break;
	}
}

#endif // VOICE_ID

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Receiver State F: Wait for unkey before recording voice ID.
//		We got sent here when a voice record was requested via
//		DTMF command.

#ifdef VOICE_ID

static void Transition_State_F(void)
{
	// Select next state = F
	SET_STATE(State_F, 'F');
}

static void State_F(MSG *pEvent)
{
	Rx_SM(pEvent);

	switch(pEvent->id)
	{
		// Here, the COR went inactive.  Done repeating.

		case EV_SQUELCHED:
		case EV_PL_UNLOCKED:
			MuteChanRx();			// Turn the channel off.

			// Turn transmitter off.

			DTMF_Disable();			// Disable DTMF detection
			Ex_SendMsg(SM_TX, MSG_TX_ENDRPT);
			Transition_State_B();	// Wait for next COR.
			break;

		default:
			break;
	}
}

#endif // VOICE_ID
